# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
import abc
from typing import Any, Callable, List, Tuple, Optional, Protocol

import torch
from .constants import DeviceType
from .utils import check_jacobians_list

# There are four functions associated with each Lie group operator xxx.
# ----------------------------------------------------------------------------------
# _xxx_impl: analytic implementation of the operator, return xxx
# _jxxx_impl: analytic implementation of the operator jacobian, return jxxx and xxx
# _xxx_autograd_fn: a torch.autograd.Function wrapper of _xxx_impl
# _jxxx_autograd_fn: simply equivalent to _jxxx_impl for now
# ----------------------------------------------------------------------------------
# Note that _jxxx_impl might not exist for some operators.


def JInverseImplFactory(module):
    def _jinverse_impl(group: torch.Tensor) -> Tuple[List[torch.Tensor], torch.Tensor]:
        return [-module._adjoint_autograd_fn(group)], module._inverse_autograd_fn(group)

    return _jinverse_impl


def LeftProjectImplFactory(module):
    def _left_project_impl(
        group: torch.Tensor, tensor: torch.Tensor, dim_out: Optional[int] = None
    ) -> torch.Tensor:
        module.check_group_tensor(group)
        module.check_left_project_tensor(tensor)
        group_inverse = module._inverse_autograd_fn(group)

        return module._project_autograd_fn(
            module._left_act_autograd_fn(group_inverse, tensor, dim_out)
        )

    return _left_project_impl


class UnaryOperator(torch.autograd.Function):
    generate_vmap_rule = True

    @classmethod
    @abc.abstractmethod
    def _forward_impl(cls, tensor: torch.Tensor) -> Any:
        pass

    @classmethod
    def forward(cls, *args):
        assert len(args) in [1, 2]
        if len(args) == 1:  # torch >= 2.0, args is (tensor,)
            output = cls._forward_impl(args[0])
        else:  # args is (ctx, tensor)
            output = cls._forward_impl(args[1])
            cls.setup_context(args[0], (args[1],), output)
        return output

    @staticmethod
    def setup_context(ctx, inputs, outputs):
        pass


class UnaryOperatorOpFnType(Protocol):
    def __call__(
        self, input: torch.Tensor, jacobians: Optional[List[torch.Tensor]] = None
    ) -> torch.Tensor:
        pass


class UnaryOperatorJOpFnType(Protocol):
    def __call__(self, input: torch.Tensor) -> Tuple[List[torch.Tensor], torch.Tensor]:
        pass


def _check_jacobians_supported(
    jop_autograd_fn: Optional[Callable],
    module_name: str,
    op_name: str,
    is_kwarg: bool = True,
):
    if jop_autograd_fn is None:
        if is_kwarg:
            msg = f"Passing jacobians= is not supported by {module_name}.{op_name}"
        else:
            msg = f"{module_name}.j{op_name} is not implemented."
        raise NotImplementedError(msg)


def UnaryOperatorFactory(
    module, op_name
) -> Tuple[UnaryOperatorOpFnType, UnaryOperatorJOpFnType]:
    # Get autograd.Function wrapper of op and its jacobian
    op_autograd_fn = getattr(module, "_" + op_name + "_autograd_fn")
    jop_autograd_fn = getattr(module, "_j" + op_name + "_autograd_fn")

    def op(
        input: torch.Tensor,
        jacobians: Optional[List[torch.Tensor]] = None,
    ) -> torch.Tensor:
        if jacobians is not None:
            _check_jacobians_supported(jop_autograd_fn, module.NAME, op_name)
            check_jacobians_list(jacobians)
            jacobians_op = jop_autograd_fn(input)[0]
            jacobians.append(jacobians_op[0])
        return op_autograd_fn(input)

    def jop(input: torch.Tensor) -> Tuple[List[torch.Tensor], torch.Tensor]:
        _check_jacobians_supported(
            jop_autograd_fn, module.name, op_name, is_kwarg=False
        )
        return jop_autograd_fn(input)

    return op, jop


class BinaryOperator(torch.autograd.Function):
    generate_vmap_rule = True

    @classmethod
    @abc.abstractmethod
    def _forward_impl(cls, input0, input1):
        pass

    @classmethod
    def forward(cls, *args):
        assert len(args) in [2, 3]
        if len(args) == 2:  # torch >= 2.0, args is (tensor1, tensor2)
            output = cls._forward_impl(args[0], args[1])
        else:  # args is (ctx, tensor)
            output = cls._forward_impl(args[1], args[2])
            cls.setup_context(args[0], (args[1], args[2]), output)
        return output

    @classmethod
    def setup_context(cls, ctx, inputs, outputs):
        pass


class BinaryOperatorOpFnType(Protocol):
    def __call__(
        self,
        input0: torch.Tensor,
        input1: torch.Tensor,
        jacobians: Optional[List[torch.Tensor]] = None,
    ) -> torch.Tensor:
        pass


class BinaryOperatorJOpFnType(Protocol):
    def __call__(
        self, input0: torch.Tensor, input1: torch.Tensor
    ) -> Tuple[List[torch.Tensor], torch.Tensor]:
        pass


def BinaryOperatorFactory(
    module, op_name
) -> Tuple[BinaryOperatorOpFnType, BinaryOperatorJOpFnType]:
    # Get autograd.Function wrapper of op and its jacobian
    op_autograd_fn = getattr(module, "_" + op_name + "_autograd_fn")
    jop_autograd_fn = getattr(module, "_j" + op_name + "_autograd_fn")

    def op(
        input0: torch.Tensor,
        input1: torch.Tensor,
        jacobians: Optional[List[torch.Tensor]] = None,
    ) -> torch.Tensor:
        if jacobians is not None:
            _check_jacobians_supported(jop_autograd_fn, module.NAME, op_name)
            check_jacobians_list(jacobians)
            jacobians_op = jop_autograd_fn(input0, input1)[0]
            for jacobian in jacobians_op:
                jacobians.append(jacobian)
        return op_autograd_fn(input0, input1)

    def jop(
        input0: torch.Tensor, input1: torch.Tensor
    ) -> Tuple[List[torch.Tensor], torch.Tensor]:
        _check_jacobians_supported(
            jop_autograd_fn, module.NAME, op_name, is_kwarg=False
        )
        return jop_autograd_fn(input0, input1)

    return op, jop


_CheckFnType = Callable[[torch.Tensor], None]


class _RandFnType(Protocol):
    def __call__(
        *size: int,
        generator: Optional[torch.Generator] = None,
        dtype: Optional[torch.dtype] = None,
        device: DeviceType = None,
        requires_grad: bool = False,
    ) -> torch.Tensor:
        pass


class _IdentityFnType(Protocol):
    def __call__(
        *size: int,
        dtype: Optional[torch.dtype] = None,
        device: DeviceType = None,
        requires_grad: bool = False,
    ) -> torch.Tensor:
        pass


class GradientOperator(torch.autograd.Function):
    generate_vmap_rule = True

    @classmethod
    @abc.abstractmethod
    def _forward_impl(cls, group, tensor, dim_out):
        pass

    @classmethod
    def forward(cls, *args):
        assert len(args) in [3, 4]
        if len(args) == 3:  # torch >= 2.0, args is (group, tensor, dim_out)
            output = cls._forward_impl(args[0], args[1], args[2])
        else:  # args is (ctx, group, tensor, dim_out)
            output = cls._forward_impl(args[1], args[2], args[3])
            cls.setup_context(args[0], (args[1], args[2], args[3]), output)
        return output

    @classmethod
    def setup_context(cls, ctx, inputs, outputs):
        pass


class GradientOperatorOpFnType(Protocol):
    def __call__(
        self,
        group: torch.Tensor,
        tensor: torch.Tensor,
        dim_out: Optional[int] = None,
    ) -> torch.Tensor:
        pass


class GradientOperatorJOpFnType(Protocol):
    def __call__(
        self,
        group: torch.Tensor,
        tensor: torch.Tensor,
        dim_out: Optional[int] = None,
    ) -> Tuple[List[torch.Tensor], torch.Tensor]:
        pass


def GradientOperatorFactory(
    module, op_name
) -> Tuple[GradientOperatorOpFnType, GradientOperatorJOpFnType]:
    # Get autograd.Function wrapper of op and its jacobian
    op_autograd_fn = getattr(module, "_" + op_name + "_autograd_fn")
    jop_autograd_fn = getattr(module, "_j" + op_name + "_autograd_fn")

    def op(
        group: torch.Tensor,
        tensor: torch.Tensor,
        dim_out: Optional[int] = None,
    ) -> torch.Tensor:
        return op_autograd_fn(group, tensor, dim_out)

    def jop(
        group: torch.Tensor,
        tensor: torch.Tensor,
        dim_out: Optional[int] = None,
    ) -> Tuple[List[torch.Tensor], torch.Tensor]:
        _check_jacobians_supported(
            jop_autograd_fn, module.NAME, op_name, is_kwarg=False
        )
        return jop_autograd_fn(group, tensor, dim_out)

    return op, jop


# Namespace to facilitate type-checking downstream
class LieGroupFns:
    def __init__(self, module):
        self.exp, self.jexp = UnaryOperatorFactory(module, "exp")
        self.log, self.jlog = UnaryOperatorFactory(module, "log")
        self.adj = UnaryOperatorFactory(module, "adjoint")[0]
        self.normalize = UnaryOperatorFactory(module, "normalize")[0]
        self.inv, self.jinv = UnaryOperatorFactory(module, "inverse")
        self.hat = UnaryOperatorFactory(module, "hat")[0]
        self.vee = UnaryOperatorFactory(module, "vee")[0]
        self.lift = UnaryOperatorFactory(module, "lift")[0]
        self.project = UnaryOperatorFactory(module, "project")[0]
        self.compose, self.jcompose = BinaryOperatorFactory(module, "compose")
        self.left_act = GradientOperatorFactory(module, "left_act")[0]
        self.left_project = GradientOperatorFactory(module, "left_project")[0]
        self.transform_from, self.jtransform_from = BinaryOperatorFactory(
            module, "transform_from"
        )
        if hasattr(module, "QuaternionToRotation"):
            (
                self.quaternion_to_rotation,
                self.jquaternion_to_rotation,
            ) = UnaryOperatorFactory(module, "quaternion_to_rotation")
        self.check_group_tensor: _CheckFnType = module.check_group_tensor
        self.check_tangent_vector: _CheckFnType = module.check_tangent_vector
        self.check_hat_tensor: _CheckFnType = module.check_hat_tensor
        if hasattr(module, "check_unit_quaternion"):
            self.check_unit_quaternion: _CheckFnType = module.check_unit_quaternion
        self.check_lift_tensor: _CheckFnType = module.check_lift_tensor
        self.check_project_tensor: _CheckFnType = module.check_project_tensor
        self.check_left_act_tensor: _CheckFnType = module.check_left_act_tensor
        self.check_left_project_tensor: _CheckFnType = module.check_left_project_tensor
        self.rand: _RandFnType = module.rand
        self.randn: _RandFnType = module.randn
        self.identity: _IdentityFnType = module.identity



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
from typing import Optional, Union
import math

import torch

TEST_EPS = 5e-7
EPS = 1e-10
PI = math.pi

__FROM_THESEUS_LAYER_TOKEN__ = "__FROM_THESEUS_LAYER_TOKEN__"


_NON_ZERO = 1.0

_INF = torch.inf

_NEAR_ZERO_D_ONE_MINUS_COSINE_BY_THETA2 = -1 / 12.0

_NEAR_ZERO_D_THETA_MINUS_SINE_BY_THETA3 = -1 / 60.0

DeviceType = Optional[Union[str, torch.device]]



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import threading
from typing import Callable

import torch


class _LieGroupCheckContext:
    contexts = threading.local()

    @classmethod
    def get_context(cls):
        if not hasattr(cls.contexts, "check_lie_group"):
            cls.contexts.check_lie_group = False
        return cls.contexts.check_lie_group

    @classmethod
    def set_context(cls, check_lie_group: bool):
        cls.contexts.check_lie_group = check_lie_group


class enable_checks(_LieGroupCheckContext):
    def __init__(self) -> None:
        pass

    def __enter__(self) -> None:
        self.prev = _LieGroupCheckContext.get_context()
        _LieGroupCheckContext.set_context(True)

    def __exit__(self, typ, value, traceback) -> None:
        _LieGroupCheckContext.set_context(self.prev)


@torch.no_grad()
def checks_base(tensor: torch.Tensor, check_impl: Callable[[torch.Tensor], None]):
    if not _LieGroupCheckContext.get_context():
        return
    if torch._C._functorch.is_batchedtensor(tensor):
        raise RuntimeError("Lie group checks must be turned off to run with vmap.")
    check_impl(tensor)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
from .check_contexts import enable_checks
from .se3_impl import _fns as SE3
from .so3_impl import _fns as SO3



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import torch
from typing import cast, List, Tuple, Optional

from torchlie.options import _TORCHLIE_GLOBAL_OPTIONS as LIE_OPTS
from . import constants, lie_group, so3_impl as SO3
from .check_contexts import checks_base
from .utils import get_module, shape_err_msg


NAME: str = "SE3"
DIM: int = 6
_DIAG_6_IDX = [0, 1, 2, 3, 4, 5]

_module = get_module(__name__)


def check_group_tensor(tensor: torch.Tensor):
    def _impl(t_: torch.Tensor):
        SO3.check_group_tensor(t_[..., :3])

    if tensor.shape[-2:] != (3, 4):
        raise ValueError(shape_err_msg("SE3 data tensors", "(..., 3, 4)", tensor.shape))

    checks_base(tensor, _impl)


def check_group_shape(tensor: torch.Tensor):
    if tensor.shape[-2:] != (3, 4):
        raise ValueError(shape_err_msg("SE3 data tensors", "(..., 3, 4)", tensor.shape))


def check_tangent_vector(tangent_vector: torch.Tensor):
    _check = tangent_vector.shape[-1] == 6
    if not _check:
        raise ValueError(
            shape_err_msg(
                "Tangent vectors of SE3",
                "(..., 6)",
                tangent_vector.shape,
            )
        )


def check_hat_tensor(tensor: torch.Tensor):
    def _impl(t_: torch.Tensor):
        if t_[..., -1].abs().max() > LIE_OPTS.get_eps("se3", "hat", t_.dtype):
            raise ValueError("The last row for hat tensors of SE3 must be zero")

        SO3.check_hat_tensor(t_[..., :3, :3])

    if tensor.shape[-2:] != (4, 4):
        raise ValueError(
            shape_err_msg("Hat tensors of SE3", "(..., 4, 4)", tensor.shape)
        )

    checks_base(tensor, _impl)


def check_transform_tensor(tensor: torch.Tensor):
    if tensor.shape[-1] != 3:
        raise ValueError(
            shape_err_msg(
                "Tensors transformed by SE3",
                "(..., 3)",
                tensor.shape,
            )
        )


def check_lift_tensor(tensor: torch.Tensor):
    if not tensor.shape[-1] == 6:
        raise ValueError(
            shape_err_msg("Lifted tensors of SE3", "(..., 6)", tensor.shape)
        )


def check_project_tensor(tensor: torch.Tensor):
    if not tensor.shape[-2:] == (3, 4):
        raise ValueError(
            shape_err_msg("Projected tensors of SE3", "(..., 3, 4)", tensor.shape)
        )


def check_left_act_tensor(tensor: torch.Tensor):
    if tensor.shape[-2] != 3:
        raise ValueError(
            shape_err_msg("Left acted tensors of SE3", "(..., 3, -1)", tensor.shape)
        )


def check_left_project_tensor(tensor: torch.Tensor):
    if tensor.shape[-2:] != (3, 4):
        raise ValueError(
            shape_err_msg("Left projected matrices of SE3", "(..., 3, 4)", tensor.shape)
        )


def get_group_size(group: torch.Tensor):
    return group.shape[:-2]


def get_tangent_vector_size(tangent_vector: torch.Tensor):
    return tangent_vector.shape[:-1]


def get_transform_tensor_size(transform_tensor: torch.Tensor):
    return transform_tensor.shape[:-1]


# -----------------------------------------------------------------------------
# Rand
# -----------------------------------------------------------------------------
# TODO: Remove duplicate code between rand and randn
def rand(
    *size: int,
    generator: Optional[torch.Generator] = None,
    dtype: Optional[torch.dtype] = None,
    device: constants.DeviceType = None,
    requires_grad: bool = False,
) -> torch.Tensor:
    rotation = SO3.rand(*size, generator=generator, dtype=dtype, device=device)
    translation = torch.rand(
        *size, 3, 1, generator=generator, dtype=dtype, device=device
    )
    ret = torch.cat((rotation, translation), dim=-1)
    ret.requires_grad_(requires_grad)
    return ret


# -----------------------------------------------------------------------------
# Randn
# -----------------------------------------------------------------------------
def randn(
    *size: int,
    generator: Optional[torch.Generator] = None,
    dtype: Optional[torch.dtype] = None,
    device: constants.DeviceType = None,
    requires_grad: bool = False,
) -> torch.Tensor:
    rotation = SO3.randn(*size, generator=generator, dtype=dtype, device=device)
    translation = torch.randn(
        *size, 3, 1, generator=generator, dtype=dtype, device=device
    )
    ret = torch.cat((rotation, translation), dim=-1)
    ret.requires_grad_(requires_grad)
    return ret


# -----------------------------------------------------------------------------
# Identity
# -----------------------------------------------------------------------------
_BASE_IDENTITY_SE3 = [[1.0, 0.0, 0.0, 0.0], [0.0, 1.0, 0.0, 0.0], [0.0, 0.0, 1.0, 0.0]]


def identity(
    *size: int,
    dtype: Optional[torch.dtype] = None,
    device: constants.DeviceType = None,
    requires_grad: bool = False,
) -> torch.Tensor:
    ret = torch.tensor(_BASE_IDENTITY_SE3, dtype=dtype, device=device).repeat(
        *size, 1, 1
    )
    ret.requires_grad_(requires_grad)
    return ret


# -----------------------------------------------------------------------------
# Exponential Map
# -----------------------------------------------------------------------------
def _exp_impl_helper(tangent_vector: torch.Tensor):
    size = get_tangent_vector_size(tangent_vector)
    ret = tangent_vector.new_zeros(*size, 3, 4)

    # Compute the rotation
    ret[..., :3], (
        theta,
        theta2,
        theta_nz,
        theta2_nz,
        sine,
        _,
        sine_by_theta,
        one_minus_cosine_by_theta2,
    ) = SO3._exp_impl_helper(tangent_vector[..., 3:])

    # Compute the translation
    tangent_vector_lin = tangent_vector[..., :3].view(*size, 3, 1)
    tangent_vector_ang = tangent_vector[..., 3:].view(*size, 3, 1)
    near_zero = theta < LIE_OPTS.get_eps("so3", "near_zero", tangent_vector.dtype)
    theta3_nz = theta_nz * theta2_nz
    theta_minus_sine_by_theta3_t = torch.where(
        near_zero, 1.0 / 6 - theta2 / 120, (theta - sine) / theta3_nz
    )
    ret[..., 3:] = sine_by_theta * tangent_vector_lin
    ret[..., 3:] += one_minus_cosine_by_theta2 * torch.cross(
        tangent_vector_ang, tangent_vector_lin, dim=-2
    )
    ret[..., 3:] += theta_minus_sine_by_theta3_t * (
        tangent_vector_ang @ (tangent_vector_ang.transpose(-1, -2) @ tangent_vector_lin)
    )

    return ret, (
        theta,
        theta2_nz,
        sine_by_theta,
        one_minus_cosine_by_theta2,
        theta_minus_sine_by_theta3_t,
    )


def _exp_impl(tangent_vector: torch.Tensor) -> torch.Tensor:
    check_tangent_vector(tangent_vector)
    ret, _ = _exp_impl_helper(tangent_vector)
    return ret


def _jexp_impl_helper(
    tangent_vector: torch.Tensor,
    rotation: torch.Tensor,
    theta: torch.Tensor,
    theta2_nz: torch.Tensor,
    sine_by_theta: torch.Tensor,
    one_minus_cosine_by_theta2: torch.Tensor,
    theta_minus_sine_by_theta3_t: torch.Tensor,
    theta_minus_sine_by_theta3_rot: torch.Tensor,
):
    size = get_tangent_vector_size(tangent_vector)
    jac = tangent_vector.new_zeros(*size, 6, 6)

    # compute rotation jacobians
    jac[..., :3, :3], _ = SO3._jexp_impl_helper(
        tangent_vector[..., 3:],
        sine_by_theta,
        one_minus_cosine_by_theta2,
        theta_minus_sine_by_theta3_rot,
    )
    jac[..., 3:, 3:] = jac[..., :3, :3]

    # compute translation jacobians
    near_zero = theta < LIE_OPTS.get_eps("so3", "near_zero", tangent_vector.dtype)
    d_one_minus_cosine_by_theta2 = torch.where(
        near_zero,
        constants._NEAR_ZERO_D_ONE_MINUS_COSINE_BY_THETA2,
        (sine_by_theta - 2 * one_minus_cosine_by_theta2) / theta2_nz,
    )
    d_theta_minus_sine_by_theta3 = torch.where(
        near_zero,
        constants._NEAR_ZERO_D_THETA_MINUS_SINE_BY_THETA3,
        (one_minus_cosine_by_theta2 - 3 * theta_minus_sine_by_theta3_t) / theta2_nz,
    )

    w = tangent_vector[..., 3:]
    v = tangent_vector[..., :3]
    wv = w.cross(v, dim=-1)
    wwv = w.cross(wv, dim=-1)
    sw = theta_minus_sine_by_theta3_t.view(*size, 1) * w

    jac_temp_t = (
        d_one_minus_cosine_by_theta2.view(*size, 1) * wv
        + d_theta_minus_sine_by_theta3.view(*size, 1) * wwv
    ).view(*size, 3, 1) @ w.view(*size, 1, 3)
    jac_temp_t -= v.view(*size, 3, 1) @ sw.view(*size, 1, 3)
    jac_temp_v = (
        -one_minus_cosine_by_theta2.view(*size, 1) * v
        - theta_minus_sine_by_theta3_t.view(*size, 1) * wv
    )
    jac_temp_t += SO3._hat_autograd_fn(jac_temp_v)
    diag_jac_t = torch.diagonal(jac_temp_t, dim1=-1, dim2=-2)
    diag_jac_t += (sw.view(*size, 1, 3) @ v.view(*size, 3, 1)).view(*size, 1)

    jac[..., :3, 3:] = rotation.transpose(-1, -2) @ jac_temp_t

    return jac, (None,)


def _jexp_impl(
    tangent_vector: torch.Tensor,
) -> Tuple[List[torch.Tensor], torch.Tensor]:
    check_tangent_vector(tangent_vector)
    ret, (
        theta,
        theta2_nz,
        sine_by_theta,
        one_minus_cosine_by_theta2,
        theta_minus_sine_by_theta3_t,
    ) = _exp_impl_helper(tangent_vector)
    near_zero = theta < LIE_OPTS.get_eps("so3", "near_zero", tangent_vector.dtype)
    theta_minus_sine_by_theta3_rot = torch.where(
        near_zero, torch.zeros_like(theta), theta_minus_sine_by_theta3_t
    )
    jac, _ = _jexp_impl_helper(
        tangent_vector,
        ret[..., :3],
        theta,
        theta2_nz,
        sine_by_theta,
        one_minus_cosine_by_theta2,
        theta_minus_sine_by_theta3_t,
        theta_minus_sine_by_theta3_rot,
    )

    return [jac], ret


class Exp(lie_group.UnaryOperator):
    @classmethod
    def _forward_impl(cls, tangent_vector):
        tangent_vector: torch.Tensor = cast(torch.Tensor, tangent_vector)
        ret = _exp_impl(tangent_vector)
        return ret

    @classmethod
    def setup_context(cls, ctx, inputs, outputs):
        # inputs is (tangent_vector,). outputs is exp_map
        ctx.save_for_backward(inputs[0], outputs)

    @classmethod
    def backward(cls, ctx, grad_output):
        tangent_vector: torch.Tensor = ctx.saved_tensors[0]
        group: torch.Tensor = ctx.saved_tensors[1]
        jacs = _jexp_impl(tangent_vector)[0][0]
        size = get_tangent_vector_size(tangent_vector)
        dg = group[..., :3].transpose(-2, -1) @ grad_output
        grad_input = jacs.transpose(-2, -1) @ torch.stack(
            (
                dg[..., 0, 3],
                dg[..., 1, 3],
                dg[..., 2, 3],
                dg[..., 2, 1] - dg[..., 1, 2],
                dg[..., 0, 2] - dg[..., 2, 0],
                dg[..., 1, 0] - dg[..., 0, 1],
            ),
            dim=-1,
        ).view(*size, 6, 1)
        return grad_input.view_as(tangent_vector)


# TODO: Implement analytic backward for _jexp_impl
_exp_autograd_fn = Exp.apply
_jexp_autograd_fn = _jexp_impl


# -----------------------------------------------------------------------------
# Logarithm Map
# -----------------------------------------------------------------------------
def _log_impl_helper(group: torch.Tensor):
    check_group_tensor(group)
    size = get_group_size(group)

    # Compute the rotation
    ret_ang, (theta, sine, cosine) = SO3._log_impl_helper(group[..., :3])

    # Compute the translation
    near_zero = theta < LIE_OPTS.get_eps("so3", "near_zero", group.dtype)
    theta2 = theta**2
    sine_theta = sine * theta
    two_cosine_minus_two = 2 * cosine - 2
    two_cosine_minus_two_nz = torch.where(
        near_zero, constants._NON_ZERO, two_cosine_minus_two
    )

    theta2_nz = torch.where(near_zero, constants._NON_ZERO, theta2)

    a = torch.where(near_zero, 1 - theta2 / 12, -sine_theta / two_cosine_minus_two_nz)
    b = torch.where(
        near_zero,
        1.0 / 12 + theta2 / 720,
        (sine_theta + two_cosine_minus_two) / (theta2_nz * two_cosine_minus_two_nz),
    )

    translation = group[..., 3].view(*size, 3, 1)
    ret_lin = a.view(*size, 1) * group[..., 3]
    ret_lin -= 0.5 * torch.cross(ret_ang, group[..., 3], dim=-1)
    ret_ang_ext = ret_ang.view(*size, 3, 1)
    ret_lin += b.view(*size, 1) * (
        ret_ang_ext @ (ret_ang_ext.transpose(-1, -2) @ translation)
    ).view(*size, 3)

    return torch.cat([ret_lin, ret_ang], dim=-1), (
        theta,
        theta2,
        theta2_nz,
        sine,
        cosine,
        two_cosine_minus_two_nz,
        a,
        b,
    )


def _log_impl(group: torch.Tensor) -> torch.Tensor:
    check_group_tensor(group)
    ret, _ = _log_impl_helper(group)
    return ret


def _jlog_impl_helper(
    tangent_vector: torch.Tensor,
    theta: torch.Tensor,
    theta2: torch.Tensor,
    theta2_nz: torch.Tensor,
    sine: torch.Tensor,
    cosine: torch.Tensor,
    two_cosine_minus_two_nz: torch.Tensor,
    a: torch.Tensor,
    b: torch.Tensor,
):
    ret_lin = tangent_vector[..., :3]
    ret_ang = tangent_vector[..., 3:]
    size = get_tangent_vector_size(tangent_vector)
    d_near_zero = theta < LIE_OPTS.get_eps("so3", "d_near_zero", tangent_vector.dtype)
    jac = tangent_vector.new_zeros(*size, 6, 6)
    jac[..., :3, :3], (b_ret_ang,) = SO3._jlog_impl_helper(ret_ang, theta, sine, cosine)
    jac[..., 3:, 3:] = jac[..., :3, :3]

    theta_nz = torch.where(d_near_zero, constants._NON_ZERO, theta)
    theta4_nz = theta2_nz**2
    c = torch.where(
        d_near_zero,
        -1 / 360.0 - theta2 / 7560.0,
        -(2 * two_cosine_minus_two_nz + theta * sine + theta2)
        / (theta4_nz * two_cosine_minus_two_nz),
    )
    d = torch.where(
        d_near_zero,
        -1 / 6.0 - theta2 / 180.0,
        (theta - sine) / (theta_nz * two_cosine_minus_two_nz),
    )

    e = ret_ang.view(*size, 1, 3) @ ret_lin.view(*size, 3, 1)
    e = e.view(*size) if len(size) > 0 else e.squeeze()

    ce_ret_ang = (c * e).view(*size, 1) * ret_ang
    jac[..., :3, 3:] = ce_ret_ang.view(*size, 3, 1) * ret_ang.view(*size, 1, 3)
    jac[..., :3, 3:] += b_ret_ang.view(*size, 3, 1) * ret_lin.view(
        *size, 1, 3
    ) + ret_lin.view(*size, 3, 1) * b_ret_ang.view(*size, 1, 3)
    diag_jac_t = torch.diagonal(jac[..., :3, 3:], dim1=-1, dim2=-2)
    diag_jac_t += (e * d).view(*size, 1)

    half_ret_lin = 0.5 * ret_lin
    jac[..., 0, 4] -= half_ret_lin[..., 2]
    jac[..., 1, 3] += half_ret_lin[..., 2]
    jac[..., 0, 5] += half_ret_lin[..., 1]
    jac[..., 2, 3] -= half_ret_lin[..., 1]
    jac[..., 1, 5] -= half_ret_lin[..., 0]
    jac[..., 2, 4] += half_ret_lin[..., 0]

    return jac, (None,)


def _jlog_impl(group: torch.Tensor) -> Tuple[List[torch.Tensor], torch.Tensor]:
    check_group_tensor(group)
    tangent_vector, (
        theta,
        theta2,
        theta2_nz,
        sine,
        cosine,
        two_cosine_minus_two_nz,
        a,
        b,
    ) = _log_impl_helper(group)
    jac, _ = _jlog_impl_helper(
        tangent_vector,
        theta,
        theta2,
        theta2_nz,
        sine,
        cosine,
        two_cosine_minus_two_nz,
        a,
        b,
    )

    return [jac], tangent_vector


class Log(lie_group.UnaryOperator):
    @classmethod
    def _forward_impl(cls, group):
        group: torch.Tensor = cast(torch.Tensor, group)
        tangent_vector = _log_impl(group)
        return tangent_vector

    @classmethod
    def setup_context(cls, ctx, inputs, outputs):
        # inputs is (group,). outputs is tangent_vector
        ctx.save_for_backward(outputs, inputs[0])

    @classmethod
    def backward(cls, ctx, grad_output):
        group: torch.Tensor = ctx.saved_tensors[1]
        jacobians = _jlog_impl(group)[0][0]
        jacobians[..., 3:] *= 0.5
        temp: torch.Tensor = lift(
            (jacobians.transpose(-1, -2) @ grad_output.unsqueeze(-1)).squeeze(-1)
        )
        return group[..., :3] @ temp


# TODO: Implement analytic backward for _jlog_impl
_log_autograd_fn = Log.apply
_jlog_autograd_fn = _jlog_impl


# -----------------------------------------------------------------------------
# Adjoint Transformation
# -----------------------------------------------------------------------------
def _adjoint_impl(group: torch.Tensor) -> torch.Tensor:
    check_group_tensor(group)
    size = get_group_size(group)
    ret = group.new_zeros(*size, 6, 6)
    ret[..., :3, :3] = group[..., :3, :3]
    ret[..., 3:, 3:] = group[..., :3, :3]
    ret[..., :3, 3:] = SO3._hat_impl(group[..., :3, 3]) @ group[..., :3, :3]
    return ret


# NOTE: No jacobian is defined for the adjoint transformation
_jadjoint_impl = None


class Adjoint(lie_group.UnaryOperator):
    @classmethod
    def _forward_impl(cls, group):
        group: torch.Tensor = cast(torch.Tensor, group)
        return _adjoint_impl(group)

    @classmethod
    def setup_context(cls, ctx, inputs, outputs):
        ctx.save_for_backward(inputs[0])  # inputs is (group,)

    @classmethod
    def backward(cls, ctx, grad_output):
        group: torch.Tensor = ctx.saved_tensors[0]
        size = get_group_size(group)
        grad_input_rot = (
            grad_output[..., :3, :3]
            + grad_output[..., 3:, 3:]
            - SO3._hat_impl(group[..., 3]) @ grad_output[..., :3, 3:]
        )
        grad_input_t = SO3._project_impl(
            grad_output[..., :3, 3:] @ group[..., :3].transpose(-2, -1)
        ).view(*size, 3, 1)

        return torch.cat((grad_input_rot, grad_input_t), dim=-1)


_adjoint_autograd_fn = Adjoint.apply
_jadjoint_autograd_fn = None


# -----------------------------------------------------------------------------
# Inverse
# -----------------------------------------------------------------------------
def _inverse_impl(group: torch.Tensor) -> torch.Tensor:
    check_group_tensor(group)
    R = group[..., :3].transpose(-2, -1)
    return torch.cat((R, -R @ group[..., 3:]), dim=-1)


_jinverse_impl = lie_group.JInverseImplFactory(_module)


class Inverse(lie_group.UnaryOperator):
    @classmethod
    def _forward_impl(cls, group):
        group: torch.Tensor = cast(torch.Tensor, group)
        return _inverse_impl(group)

    @classmethod
    def setup_context(cls, ctx, inputs, outputs):
        ctx.save_for_backward(inputs[0])  # inputs is (group,)

    @classmethod
    def backward(cls, ctx, grad_output):
        group: torch.Tensor = ctx.saved_tensors[0]
        grad_input_rot = grad_output[..., :3].transpose(-1, -2) - group[
            ..., 3:
        ] @ grad_output[..., 3:].transpose(-1, -2)
        grad_input_t = -group[..., :3] @ grad_output[..., 3:]
        return torch.cat((grad_input_rot, grad_input_t), dim=-1)


_inverse_autograd_fn = Inverse.apply
_jinverse_autograd_fn = _jinverse_impl


# -----------------------------------------------------------------------------
# Hat
# -----------------------------------------------------------------------------
def _hat_impl(tangent_vector: torch.Tensor) -> torch.Tensor:
    check_tangent_vector(tangent_vector)
    size = get_tangent_vector_size(tangent_vector)
    tangent_vector = tangent_vector.view(*size, 6)
    tensor = tangent_vector.new_zeros(*size, 4, 4)
    tensor[..., :3, :3] = SO3._hat_impl(tangent_vector[..., 3:])
    tensor[..., :3, 3] = tangent_vector[..., :3]

    return tensor


# NOTE: No jacobian is defined for the hat operator
_jhat_impl = None


class Hat(lie_group.UnaryOperator):
    @classmethod
    def _forward_impl(cls, tangent_vector):
        tangent_vector: torch.Tensor = cast(torch.Tensor, tangent_vector)
        ret = _hat_impl(tangent_vector)
        return ret

    @classmethod
    def backward(cls, ctx, grad_output):
        grad_output: torch.Tensor = cast(torch.Tensor, grad_output)
        return torch.stack(
            (
                grad_output[..., 0, 3],
                grad_output[..., 1, 3],
                grad_output[..., 2, 3],
                grad_output[..., 2, 1] - grad_output[..., 1, 2],
                grad_output[..., 0, 2] - grad_output[..., 2, 0],
                grad_output[..., 1, 0] - grad_output[..., 0, 1],
            ),
            dim=-1,
        )


_hat_autograd_fn = Hat.apply
_jhat_autograd_fn = None


# -----------------------------------------------------------------------------
# Vee
# -----------------------------------------------------------------------------
def _vee_impl(tensor: torch.Tensor) -> torch.Tensor:
    check_hat_tensor(tensor)
    size = tensor.shape[:-2]
    ret = tensor.new_zeros(*size, 6)
    ret[..., :3] = tensor[..., :3, 3]
    ret[..., 3:] = 0.5 * torch.stack(
        (
            tensor[..., 2, 1] - tensor[..., 1, 2],
            tensor[..., 0, 2] - tensor[..., 2, 0],
            tensor[..., 1, 0] - tensor[..., 0, 1],
        ),
        dim=-1,
    )
    return ret


# NOTE: No jacobian is defined for the vee operator
_jvee_impl = None


class Vee(lie_group.UnaryOperator):
    @classmethod
    def _forward_impl(cls, tangent_vector):
        tangent_vector: torch.Tensor = cast(torch.Tensor, tangent_vector)
        ret = _vee_impl(tangent_vector)
        return ret

    @classmethod
    def backward(cls, ctx, grad_output):
        grad_output: torch.Tensor = cast(torch.Tensor, grad_output)
        size = grad_output.shape[:-1]
        grad_input = grad_output.new_zeros(*size, 4, 4)
        grad_input[..., :3, 3] = grad_output[..., :3]
        grad_input[..., :3, :3] = 0.5 * SO3._hat_impl(grad_output[..., 3:])
        return grad_input


_vee_autograd_fn = Vee.apply
_jvee_autograd_fn = None


# -----------------------------------------------------------------------------
# Compose
# -----------------------------------------------------------------------------
def _compose_impl(group0: torch.Tensor, group1: torch.Tensor) -> torch.Tensor:
    check_group_tensor(group0)
    check_group_tensor(group1)
    ret_rot = group0[..., :3] @ group1[..., :3]
    ret_t = group0[..., :3] @ group1[..., 3:] + group0[..., 3:]
    return torch.cat((ret_rot, ret_t), dim=-1)


def _jcompose_impl(
    group0: torch.Tensor, group1: torch.Tensor
) -> Tuple[List[torch.Tensor], torch.Tensor]:
    check_group_tensor(group0)
    check_group_tensor(group1)
    ret = _compose_impl(group0, group1)
    size = get_group_size(ret)
    jacobians = []
    jacobians.append(
        _adjoint_autograd_fn(_inverse_autograd_fn(group1)).expand(*size, 6, 6).clone()
    )
    jacobians.append(group0.new_zeros(*size, 6, 6))
    jacobians[1][..., _DIAG_6_IDX, _DIAG_6_IDX] = 1
    return jacobians, ret


class Compose(lie_group.BinaryOperator):
    @classmethod
    def _forward_impl(cls, group0, group1):
        group0: torch.Tensor = cast(torch.Tensor, group0)
        group1: torch.Tensor = cast(torch.Tensor, group1)
        ret = _compose_impl(group0, group1)
        return ret

    @classmethod
    def setup_context(cls, ctx, inputs, outputs):
        ctx.save_for_backward(inputs[0], inputs[1])  # the two groups

    @classmethod
    def backward(cls, ctx, grad_output):
        group0: torch.Tensor = ctx.saved_tensors[0]
        group1: torch.Tensor = ctx.saved_tensors[1]
        grad_input0 = torch.cat(
            (grad_output @ group1.transpose(-1, -2), grad_output[..., 3:]), dim=-1
        )
        grad_input1 = group0[..., :3].transpose(-1, -2) @ grad_output
        return grad_input0, grad_input1


_compose_autograd_fn = Compose.apply
_jcompose_autograd_fn = _jcompose_impl


# -----------------------------------------------------------------------------
# Transform From
# -----------------------------------------------------------------------------
def _transform_from_impl(group: torch.Tensor, tensor: torch.Tensor) -> torch.Tensor:
    check_group_tensor(group)
    check_transform_tensor(tensor)
    ret = group[..., -1:] + group[..., :3] @ tensor.unsqueeze(-1)
    return ret.squeeze(-1)


def _jtransform_from_impl(
    group: torch.Tensor, tensor: torch.Tensor
) -> Tuple[List[torch.Tensor], torch.Tensor]:
    check_group_tensor(group)
    check_transform_tensor(tensor)
    ret = _transform_from_impl(group, tensor)
    size = get_transform_tensor_size(ret)
    jacobian_g = group.new_empty(*size, 3, 6)
    jacobian_g[..., :3] = group[..., :3]
    jacobian_g[..., 3:] = -group[..., :3] @ SO3._hat_autograd_fn(tensor)
    jacobian_p = group[..., :3]
    jacobian_g = jacobian_g.expand(*size, 3, 6).clone()
    jacobian_p = jacobian_p.expand(*size, 3, 3).clone()
    return [jacobian_g, jacobian_p], ret


class TransformFrom(lie_group.BinaryOperator):
    @classmethod
    def _forward_impl(cls, group, tensor):
        group: torch.Tensor = cast(torch.Tensor, group)
        tensor: torch.Tensor = cast(torch.Tensor, tensor)
        ret = _transform_from_impl(group, tensor)
        return ret

    @classmethod
    def setup_context(cls, ctx, inputs, outputs):
        # inputs is (group, tensor)
        ctx.save_for_backward(inputs[0], inputs[1])

    @classmethod
    def backward(cls, ctx, grad_output):
        group: torch.Tensor = ctx.saved_tensors[0]
        tensor: torch.Tensor = ctx.saved_tensors[1]
        grad_output: torch.Tensor = grad_output.unsqueeze(-1)
        tensor_size = get_transform_tensor_size(tensor)
        grad_input0 = torch.cat(
            (grad_output @ tensor.view(*tensor_size, 1, 3), grad_output), dim=-1
        )
        grad_input1 = group[..., :3].transpose(-1, -2) @ grad_output
        return grad_input0, grad_input1.squeeze(-1)


_transform_from_autograd_fn = TransformFrom.apply
_jtransform_from_autograd_fn = _jtransform_from_impl


# -----------------------------------------------------------------------------
# Lift
# -----------------------------------------------------------------------------
def _lift_impl(tensor: torch.Tensor) -> torch.Tensor:
    check_lift_tensor(tensor)
    ret = tensor.new_zeros(tensor.shape[:-1] + (3, 4))
    ret[..., :3] = SO3._lift_impl(tensor[..., 3:])
    ret[..., 3] = tensor[..., :3]

    return ret


# NOTE: No jacobian is defined for the project operator
_jlift_impl = None


class Lift(lie_group.UnaryOperator):
    @classmethod
    def _forward_impl(cls, tensor):
        tensor: torch.Tensor = cast(torch.Tensor, tensor)
        ret = _lift_impl(tensor)
        return ret

    @classmethod
    def backward(cls, ctx, grad_output):
        grad_output: torch.Tensor = cast(torch.Tensor, grad_output)
        return project(grad_output)


_lift_autograd_fn = Lift.apply
_jlift_autograd_fn = None

lift, jlift = lie_group.UnaryOperatorFactory(_module, "lift")


# -----------------------------------------------------------------------------
# Project
# -----------------------------------------------------------------------------
def _project_impl(tensor: torch.Tensor) -> torch.Tensor:
    check_project_tensor(tensor)
    return torch.stack(
        (
            tensor[..., 0, 3],
            tensor[..., 1, 3],
            tensor[..., 2, 3],
            tensor[..., 2, 1] - tensor[..., 1, 2],
            tensor[..., 0, 2] - tensor[..., 2, 0],
            tensor[..., 1, 0] - tensor[..., 0, 1],
        ),
        dim=-1,
    )


# NOTE: No jacobian is defined for the project operator
_jproject_impl = None


class Project(lie_group.UnaryOperator):
    @classmethod
    def _forward_impl(cls, tensor):
        tensor: torch.Tensor = cast(torch.Tensor, tensor)
        ret = _project_impl(tensor)
        return ret

    @classmethod
    def backward(cls, ctx, grad_output):
        grad_output: torch.Tensor = cast(torch.Tensor, grad_output)
        return lift(grad_output)


_project_autograd_fn = Project.apply
_jproject_autograd_fn = None

project, jproject = lie_group.UnaryOperatorFactory(_module, "project")


# -----------------------------------------------------------------------------
# Left Act
# -----------------------------------------------------------------------------
def _left_act_impl(
    group: torch.Tensor, tensor: torch.Tensor, dim_out: Optional[int] = None
) -> torch.Tensor:
    check_group_tensor(group)
    check_left_act_tensor(tensor)
    dim_out = tensor.ndim - group.ndim if dim_out is None else dim_out
    ret = SO3._left_act_impl(group[..., :3], tensor, dim_out)
    return ret


def _left_act_backward_helper(
    group: torch.Tensor, tensor: torch.Tensor, dim_out: int, grad_output: torch.Tensor
):
    jac_rot, jac_tensor, _ = SO3._left_act_backward_helper(
        group[..., :3], tensor, dim_out, grad_output
    )
    return (
        torch.cat((jac_rot, jac_rot.new_zeros(*jac_rot.shape[:-2], 3, 1)), dim=-1),
        jac_tensor,
        None,
    )


class LeftAct(lie_group.GradientOperator):
    @classmethod
    def _forward_impl(cls, group, tensor, dim_out):
        group: torch.Tensor = cast(torch.Tensor, group)
        tensor: torch.Tensor = cast(torch.Tensor, tensor)
        ret = _left_act_impl(group, tensor, dim_out)
        return ret

    @classmethod
    def setup_context(cls, ctx, inputs, outputs):
        ctx.save_for_backward(inputs[0], inputs[1])
        ctx.dim_out = inputs[2]

    @classmethod
    def backward(cls, ctx, grad_output):
        group, tensor = ctx.saved_tensors
        dim_out = ctx.dim_out
        dim_out: int = tensor.ndim - group.ndim if dim_out is None else dim_out
        return _left_act_backward_helper(group, tensor, dim_out, grad_output)


def _left_act_autograd_fn(
    group: torch.Tensor, tensor: torch.Tensor, dim_out: Optional[int] = None
):
    return LeftAct.apply(group, tensor, dim_out)


_jleft_act_autograd_fn = None

left_act, jleft_act = lie_group.BinaryOperatorFactory(_module, "left_act")


# -----------------------------------------------------------------------------
# Left Project
# -----------------------------------------------------------------------------
_left_project_impl = lie_group.LeftProjectImplFactory(_module)
_jleft_project_impl = None


def _left_project_backward_helper(
    group: torch.Tensor,
    tensor: torch.Tensor,
    dim_out: int,
    grad_output_lifted: torch.Tensor,
):
    jac_rot, jac_tensor, _ = SO3._left_project_backward_helper(
        group[..., :3], tensor, dim_out, grad_output_lifted
    )
    return (
        torch.cat((jac_rot, jac_rot.new_zeros(*jac_rot.shape[:-2], 3, 1)), dim=-1),
        jac_tensor,
        None,
    )


class LeftProject(lie_group.GradientOperator):
    @classmethod
    def _forward_impl(cls, group, tensor, dim_out):
        group: torch.Tensor = cast(torch.Tensor, group)
        tensor: torch.Tensor = cast(torch.Tensor, tensor)
        ret = _left_project_impl(group, tensor, dim_out)
        return ret

    @classmethod
    def setup_context(cls, ctx, inputs, outputs):
        ctx.save_for_backward(inputs[0], inputs[1])
        ctx.dim_out = inputs[2]

    @classmethod
    def backward(cls, ctx, grad_output):
        group, tensor = ctx.saved_tensors
        dim_out: int = ctx.dim_out
        dim_out: int = tensor.ndim - group.ndim if dim_out is None else dim_out
        grad_output_lifted = lift(grad_output)
        return _left_project_backward_helper(group, tensor, dim_out, grad_output_lifted)


def _left_project_autograd_fn(
    group: torch.Tensor, tensor: torch.Tensor, dim_out: Optional[int] = None
):
    return LeftProject.apply(group, tensor, dim_out)


_jleft_project_autograd_fn = _jleft_project_impl


# -----------------------------------------------------------------------------
# Normalize
# -----------------------------------------------------------------------------
def _normalize_impl(tensor: torch.Tensor) -> torch.Tensor:
    check_group_shape(tensor)
    rotation = SO3._normalize_impl_helper(tensor[..., :3])[0]
    translation = tensor[..., 3:]
    return torch.cat((rotation, translation), dim=-1)


class Normalize(lie_group.UnaryOperator):
    @classmethod
    def _forward_impl(cls, tensor):
        check_group_shape(tensor)
        tensor: torch.Tensor = tensor
        rotation, svd_info = SO3._normalize_impl_helper(tensor[..., :3])
        translation = tensor[..., 3:]
        output = torch.cat((rotation, translation), dim=-1)
        return output, svd_info

    @staticmethod
    def setup_context(ctx, inputs, outputs):
        # outputs is (normalized_out, svd_info)
        svd_info = outputs[1]
        ctx.save_for_backward(
            svd_info["u"], svd_info["s"], svd_info["v"], svd_info["sign"]
        )

    @classmethod
    def backward(cls, ctx, grad_output, _):
        u, s, v, sign = ctx.saved_tensors
        grad_input1 = SO3._normalize_backward_helper(
            u, s, v, sign, grad_output[..., :3]
        )
        grad_input2 = grad_output[..., 3:]
        grad_input = torch.cat((grad_input1, grad_input2), dim=-1)
        return grad_input, None


def _normalize_autograd_fn(tensor: torch.Tensor):
    return Normalize.apply(tensor)[0]


_jnormalize_autograd_fn = None


_fns = lie_group.LieGroupFns(_module)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
from enum import Enum
from typing import Any, Callable, List, Optional, Protocol, Tuple, Union, TYPE_CHECKING

import torch
from .functional import SE3 as _se3_impl, SO3 as _so3_impl
from .functional.constants import DeviceType
from .functional.lie_group import LieGroupFns

if TYPE_CHECKING:
    from .lie_tensor import LieTensor, _LieTensorBase


# The next two are similar to the ones in functional, except they return a LieTensor.
class _RandFnType(Protocol):
    def __call__(
        self,
        *size: Any,
        generator: Optional[torch.Generator] = None,
        dtype: Optional[torch.dtype] = None,
        device: DeviceType = None,
        requires_grad: bool = False,
    ) -> "LieTensor":
        pass


class _IdentityFnType(Protocol):
    def __call__(
        self,
        *size: int,
        dtype: Optional[torch.dtype] = None,
        device: DeviceType = None,
        requires_grad: bool = False,
    ) -> "LieTensor":
        pass


Device = Union[torch.device, str, None]
TensorType = Union[torch.Tensor, "_LieTensorBase"]
_JFnReturnType = Tuple[List[torch.Tensor], TensorType]


class _ltype(Enum):
    SE3 = 0
    SO3 = 1


def _get_fn_lib(ltype: _ltype):
    return {
        _ltype.SE3: _se3_impl,
        _ltype.SO3: _so3_impl,
    }[ltype]


def _eval_op(
    fn_lib: LieGroupFns,
    op_name: str,
    input0: torch.Tensor,
    jacobians: Optional[List[torch.Tensor]] = None,
) -> torch.Tensor:
    return getattr(fn_lib, op_name)(input0, jacobians=jacobians)


class ltype:
    def __init__(self, lt: _ltype):
        self._lt = lt
        self._fn_lib = _get_fn_lib(lt)

    rand: _RandFnType
    randn: _RandFnType
    identity: _IdentityFnType
    _call_impl: Callable[[torch.Tensor], "LieTensor"]

    def __call__(self, tensor: torch.Tensor) -> "LieTensor":
        return self._call_impl(tensor)

    _create_lie_tensor: Callable[[torch.Tensor, "ltype"], "LieTensor"]

    def exp(self, tangent_vector: torch.Tensor) -> "LieTensor":
        return self._create_lie_tensor(
            _eval_op(self._fn_lib, "exp", tangent_vector), self
        )

    def jexp(self, tangent_vector: torch.Tensor) -> _JFnReturnType:
        jacs: List[torch.Tensor] = []
        op_res: TensorType = self._fn_lib.exp(tangent_vector, jacobians=jacs)
        return jacs, self._create_lie_tensor(op_res, self)

    def hat(self, tangent_vector: torch.Tensor) -> torch.Tensor:
        return _eval_op(self._fn_lib, "hat", tangent_vector)

    def vee(self, matrix: torch.Tensor) -> torch.Tensor:
        return _eval_op(self._fn_lib, "vee", matrix)

    def lift(self, matrix: torch.Tensor) -> torch.Tensor:
        return _eval_op(self._fn_lib, "lift", matrix)

    def project(self, matrix: torch.Tensor) -> torch.Tensor:
        return _eval_op(self._fn_lib, "project", matrix)

    def __str__(self) -> str:
        return self._lt.name


SE3 = ltype(_ltype.SE3)
SO3 = ltype(_ltype.SO3)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import List
import torch


def check_jacobians_list(jacobians: List[torch.Tensor]):
    if len(jacobians) != 0:
        raise ValueError("jacobians list to be populated must be empty.")


def get_module(module_name):
    return __import__(module_name, fromlist=[""])


def get_cls_module(cls):
    return get_module(cls.__module__)


def shape_err_msg(data_type: str, expected_shape: str, tensor_shape: torch.Size) -> str:
    return f"{data_type} must have shape {expected_shape} but got shape {tensor_shape}."


def fill_dims(tensor: torch.Tensor, dim: int) -> torch.Tensor:
    return tensor.view((1,) * (dim - tensor.ndim) + tensor.shape)


def permute_op_dim(dim: int, op_dim: int, group_dim: int):
    return (
        [*range(dim - op_dim - group_dim, dim - group_dim)]
        + [*range(0, dim - op_dim - group_dim)]
        + [*range(dim - group_dim, dim)]
    )


def unpermute_op_dim(dim: int, op_dim: int, group_dim: int):
    return (
        [*range(op_dim, dim - group_dim)]
        + [*range(0, op_dim)]
        + [*range(dim - group_dim, dim)]
    )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import torch
from typing import cast, List, Tuple, Optional

from torchlie.options import _TORCHLIE_GLOBAL_OPTIONS as LIE_OPTS
from . import constants, lie_group
from .check_contexts import checks_base
from .utils import (
    get_module,
    shape_err_msg,
    permute_op_dim,
    unpermute_op_dim,
    fill_dims,
)


NAME: str = "SO3"
DIM: int = 3

_DIAG_3_IDX = [0, 1, 2]

_module = get_module(__name__)


def check_group_tensor(tensor: torch.Tensor):
    def _impl(t_):
        MATRIX_EPS = LIE_OPTS.get_eps("so3", "matrix", t_.dtype)
        if t_.dtype != torch.float64:
            t_ = t_.double()

        _check = (
            torch.matmul(t_, t_.transpose(-1, -2))
            - torch.eye(3, 3, dtype=t_.dtype, device=t_.device)
        ).abs().max().item() < MATRIX_EPS
        _check &= (torch.linalg.det(t_) - 1).abs().max().item() < MATRIX_EPS

        if not _check:
            raise ValueError("Invalid data tensor for SO3.")

    if tensor.shape[-2:] != (3, 3):
        raise ValueError(shape_err_msg("SO3 data tensors", "(..., 3, 3)", tensor.shape))

    checks_base(tensor, _impl)


def check_group_shape(tensor: torch.Tensor):
    if tensor.shape[-2:] != (3, 3):
        raise ValueError(shape_err_msg("SO3 data tensors", "(..., 3, 3)", tensor.shape))


def check_tangent_vector(tangent_vector: torch.Tensor):
    _check = tangent_vector.shape[-1] == 3
    if not _check:
        raise ValueError(
            shape_err_msg(
                "Tangent vectors of SO3",
                "(..., 3)",
                tangent_vector.shape,
            )
        )


def check_hat_tensor(tensor: torch.Tensor):
    def _impl(t_: torch.Tensor):
        if (t_.transpose(-1, -2) + t_).abs().max().item() > LIE_OPTS.get_eps(
            "so3", "hat", t_.dtype
        ):
            raise ValueError("Hat tensors of SO3 can only be skew-symmetric.")

    if tensor.shape[-2:] != (3, 3):
        raise ValueError(
            shape_err_msg("Hat tensors of SO3", "(..., 3, 3)", tensor.shape)
        )

    checks_base(tensor, _impl)


def check_transform_tensor(tensor: torch.Tensor):
    if tensor.shape[-1] != 3:
        raise ValueError(
            shape_err_msg(
                "Tensors transformed by SO3",
                "(..., 3)",
                tensor.shape,
            )
        )


def check_lift_tensor(tensor: torch.Tensor):
    if not tensor.shape[-1] == 3:
        raise ValueError(
            shape_err_msg("Lifted tensors of SO3", "(..., 3)", tensor.shape)
        )


def check_project_tensor(tensor: torch.Tensor):
    if not tensor.shape[-2:] == (3, 3):
        raise ValueError(
            shape_err_msg("Projected tensors of SO3", "(..., 3, 3)", tensor.shape)
        )


def check_unit_quaternion(quaternion: torch.Tensor):
    def _impl(t_: torch.Tensor):
        QUANTERNION_EPS = LIE_OPTS.get_eps("so3", "quat", t_.dtype)

        if t_.dtype != torch.float64:
            t_ = t_.double()

        if (torch.linalg.norm(t_, dim=-1) - 1).abs().max().item() >= QUANTERNION_EPS:
            raise ValueError("Not unit quaternions.")

    if quaternion.shape[-1] != 4:
        raise ValueError(shape_err_msg("Quaternions", "(..., 4)", quaternion.shape))

    checks_base(quaternion, _impl)


def check_left_act_tensor(tensor: torch.Tensor):
    if tensor.shape[-2] != 3:
        raise ValueError(
            shape_err_msg("Left acted tensors of SO3", "(..., 3, -1)", tensor.shape)
        )


def check_left_project_tensor(tensor: torch.Tensor):
    if tensor.shape[-2:] != (3, 3):
        raise ValueError(
            shape_err_msg("Left projected tensors of SO3", "(..., 3, 3)", tensor.shape)
        )


def get_group_size(group: torch.Tensor):
    return group.shape[:-2]


def get_tangent_vector_size(tangent_vector: torch.Tensor):
    return tangent_vector.shape[:-1]


def get_transform_tensor_size(transform_tensor: torch.Tensor):
    return transform_tensor.shape[:-1]


# -----------------------------------------------------------------------------
# Rand
# -----------------------------------------------------------------------------


def rand(
    *size: int,
    generator: Optional[torch.Generator] = None,
    dtype: Optional[torch.dtype] = None,
    device: constants.DeviceType = None,
    requires_grad: bool = False,
) -> torch.Tensor:
    # Reference:
    # https://web.archive.org/web/20211105205926/http://planning.cs.uiuc.edu/node198.html
    u = torch.rand(3, *size, generator=generator, dtype=dtype, device=device)
    u1 = u[0]
    u2, u3 = u[1:3] * 2 * constants.PI

    a = torch.sqrt(1.0 - u1)
    b = torch.sqrt(u1)
    quaternion = torch.stack(
        [
            a * torch.sin(u2),
            a * torch.cos(u2),
            b * torch.sin(u3),
            b * torch.cos(u3),
        ],
        dim=-1,
    )
    assert quaternion.shape == size + (4,)
    ret = _quaternion_to_rotation_autograd_fn(quaternion)
    ret.requires_grad_(requires_grad)
    return ret


# -----------------------------------------------------------------------------
# Randn
# -----------------------------------------------------------------------------
def randn(
    *size: int,
    generator: Optional[torch.Generator] = None,
    dtype: Optional[torch.dtype] = None,
    device: constants.DeviceType = None,
    requires_grad: bool = False,
) -> torch.Tensor:
    ret = _exp_autograd_fn(
        constants.PI
        * torch.randn(*size, 3, generator=generator, dtype=dtype, device=device)
    )
    ret.requires_grad_(requires_grad)
    return ret


# -----------------------------------------------------------------------------
# Identity
# -----------------------------------------------------------------------------
def identity(
    *size: int,
    dtype: Optional[torch.dtype] = None,
    device: constants.DeviceType = None,
    requires_grad: bool = False,
) -> torch.Tensor:
    ret = torch.eye(3, device=device, dtype=dtype).repeat(*size, 1, 1)
    ret.requires_grad_(requires_grad)
    return ret


# -----------------------------------------------------------------------------
# Exponential Map
# -----------------------------------------------------------------------------
def _exp_impl_helper(tangent_vector: torch.Tensor):
    theta = torch.linalg.norm(tangent_vector, dim=-1, keepdim=True).unsqueeze(-1)
    theta2 = theta**2
    # Compute the approximations when theta ~ 0
    near_zero = theta < LIE_OPTS.get_eps("so3", "near_zero", tangent_vector.dtype)
    theta_nz = torch.where(near_zero, constants._NON_ZERO, theta)
    theta2_nz = torch.where(near_zero, constants._NON_ZERO, theta2)

    cosine = torch.where(near_zero, 8 / (4 + theta2) - 1, theta.cos())
    sine = theta.sin()
    sine_by_theta = torch.where(near_zero, 0.5 * cosine + 0.5, sine / theta_nz)
    one_minus_cosine_by_theta2 = torch.where(
        near_zero, 0.5 * sine_by_theta, (1 - cosine) / theta2_nz
    )

    size = tangent_vector.shape[:-1]
    ret = (
        one_minus_cosine_by_theta2
        * tangent_vector.view(*size, 3, 1)
        @ tangent_vector.view(*size, 1, 3)
    )
    ret[..., 0, 0] += cosine.view(size)
    ret[..., 1, 1] += cosine.view(size)
    ret[..., 2, 2] += cosine.view(size)
    sine_axis = sine_by_theta.view(*size, 1) * tangent_vector
    ret[..., 0, 1] -= sine_axis[..., 2]
    ret[..., 1, 0] += sine_axis[..., 2]
    ret[..., 0, 2] += sine_axis[..., 1]
    ret[..., 2, 0] -= sine_axis[..., 1]
    ret[..., 1, 2] -= sine_axis[..., 0]
    ret[..., 2, 1] += sine_axis[..., 0]

    return ret, (
        theta,
        theta2,
        theta_nz,
        theta2_nz,
        sine,
        cosine,
        sine_by_theta,
        one_minus_cosine_by_theta2,
    )


def _exp_impl(tangent_vector: torch.Tensor) -> torch.Tensor:
    check_tangent_vector(tangent_vector)
    ret, _ = _exp_impl_helper(tangent_vector)
    return ret


def _jexp_impl_helper(
    tangent_vector: torch.Tensor,
    sine_by_theta: torch.Tensor,
    one_minus_cosine_by_theta2: torch.Tensor,
    theta_minus_sine_by_theta3: torch.Tensor,
):
    size = tangent_vector.shape[:-1]
    jac = theta_minus_sine_by_theta3 * (
        tangent_vector.view(*size, 3, 1) @ tangent_vector.view(*size, 1, 3)
    )
    diag_jac = jac.diagonal(dim1=-1, dim2=-2)
    diag_jac += sine_by_theta.view(*size, 1)
    jac_temp = one_minus_cosine_by_theta2.view(*size, 1) * tangent_vector
    jac[..., 0, 1] += jac_temp[..., 2]
    jac[..., 1, 0] -= jac_temp[..., 2]
    jac[..., 0, 2] -= jac_temp[..., 1]
    jac[..., 2, 0] += jac_temp[..., 1]
    jac[..., 1, 2] += jac_temp[..., 0]
    jac[..., 2, 1] -= jac_temp[..., 0]

    return jac, (None,)


def _jexp_impl(
    tangent_vector: torch.Tensor,
) -> Tuple[List[torch.Tensor], torch.Tensor]:
    check_tangent_vector(tangent_vector)
    ret, (
        theta,
        _,
        theta_nz,
        theta2_nz,
        sine,
        _,
        sine_by_theta,
        one_minus_cosine_by_theta2,
    ) = _exp_impl_helper(tangent_vector)

    near_zero = theta < LIE_OPTS.get_eps("so3", "near_zero", tangent_vector.dtype)
    theta3_nz = theta_nz * theta2_nz
    theta_minus_sine_by_theta3 = torch.where(
        near_zero, torch.zeros_like(theta), (theta - sine) / theta3_nz
    )
    jac, _ = _jexp_impl_helper(
        tangent_vector,
        sine_by_theta,
        one_minus_cosine_by_theta2,
        theta_minus_sine_by_theta3,
    )

    return [jac], ret


class Exp(lie_group.UnaryOperator):
    @classmethod
    def _forward_impl(cls, tangent_vector):
        tangent_vector: torch.Tensor = cast(torch.Tensor, tangent_vector)
        ret = _exp_impl(tangent_vector)
        return ret

    @classmethod
    def setup_context(cls, ctx, inputs, outputs):
        # inputs is (tangent_vector, ). outputs is exp_map
        ctx.save_for_backward(inputs[0], outputs)

    @classmethod
    def backward(cls, ctx, grad_output):
        tangent_vector: torch.Tensor = ctx.saved_tensors[0]
        group: torch.Tensor = ctx.saved_tensors[1]
        jacs = _jexp_impl(tangent_vector)[0][0]
        dR = group.transpose(-2, -1) @ grad_output
        size = get_tangent_vector_size(tangent_vector)
        grad_input = jacs.transpose(-2, -1) @ torch.stack(
            (
                dR[..., 2, 1] - dR[..., 1, 2],
                dR[..., 0, 2] - dR[..., 2, 0],
                dR[..., 1, 0] - dR[..., 0, 1],
            ),
            dim=-1,
        ).view(*size, 3, 1)
        return grad_input.view_as(tangent_vector)


# TODO: Implement analytic backward for _jexp_impl
_exp_autograd_fn = Exp.apply
_jexp_autograd_fn = _jexp_impl


# -----------------------------------------------------------------------------
# Logarithm Map
# -----------------------------------------------------------------------------
def _log_impl_helper(group: torch.Tensor):
    size = get_group_size(group)
    sine_axis = group.new_zeros(*size, 3)
    sine_axis[..., 0] = 0.5 * (group[..., 2, 1] - group[..., 1, 2])
    sine_axis[..., 1] = 0.5 * (group[..., 0, 2] - group[..., 2, 0])
    sine_axis[..., 2] = 0.5 * (group[..., 1, 0] - group[..., 0, 1])
    cosine = 0.5 * (group.diagonal(dim1=-1, dim2=-2).sum(dim=-1) - 1)
    sine = sine_axis.norm(dim=-1)
    theta = torch.atan2(sine, cosine)

    near_zero = theta < LIE_OPTS.get_eps("so3", "near_zero", group.dtype)
    near_pi = 1 + cosine <= LIE_OPTS.get_eps("so3", "near_pi", group.dtype)
    # theta != pi
    near_zero_or_near_pi = torch.logical_or(near_zero, near_pi)
    # Compute the approximation of theta / sin(theta) when theta is near to 0
    sine_nz = torch.where(near_zero_or_near_pi, constants._NON_ZERO, sine)
    scale = torch.where(
        near_zero_or_near_pi,
        1 + sine**2 / 6,
        theta / sine_nz,
    )
    ret = sine_axis * scale.view(*size, 1)

    # # theta ~ pi
    ddiag = torch.diagonal(group, dim1=-1, dim2=-2)
    # Find the index of major columns and diagonals
    major = torch.logical_and(
        ddiag[..., 1] > ddiag[..., 0], ddiag[..., 1] > ddiag[..., 2]
    ) + 2 * torch.logical_and(
        ddiag[..., 2] > ddiag[..., 0], ddiag[..., 2] > ddiag[..., 1]
    )
    major = major.view(-1)
    aux = torch.ones(size, dtype=torch.bool)
    sel_rows = 0.5 * (group[aux, major] + group[aux, :, major]).view(*size, 3)
    sel_rows[aux, major] -= cosine.view(-1)
    axis = sel_rows / torch.where(
        near_zero,
        constants._NON_ZERO,
        sel_rows.norm(dim=-1),
    ).view(*size, 1)
    sign_tmp = sine_axis[aux, major].sign().view(size)
    sign = torch.where(sign_tmp != 0, sign_tmp, torch.ones_like(sign_tmp))
    tangent_vector = torch.where(
        near_pi.view(*size, 1), axis * (theta * sign).view(*size, 1), ret
    )

    return tangent_vector, (theta, sine, cosine)


def _log_impl(group: torch.Tensor) -> torch.Tensor:
    check_group_tensor(group)
    tangent_vector, _ = _log_impl_helper(group)
    return tangent_vector


def _jlog_impl_helper(
    tangent_vector: torch.Tensor,
    theta: torch.Tensor,
    sine: torch.Tensor,
    cosine: torch.Tensor,
):
    size = get_tangent_vector_size(tangent_vector)
    d_near_zero = theta < LIE_OPTS.get_eps("so3", "d_near_zero", tangent_vector.dtype)
    theta2 = theta**2
    sine_theta = sine * theta
    two_cosine_minus_two = 2 * cosine - 2
    two_cosine_minus_two_nz = torch.where(
        d_near_zero, constants._NON_ZERO, two_cosine_minus_two
    )
    theta2_nz = torch.where(d_near_zero, constants._NON_ZERO, theta2)

    a = torch.where(d_near_zero, 1 - theta2 / 12, -sine_theta / two_cosine_minus_two_nz)
    b = torch.where(
        d_near_zero,
        1.0 / 12 + theta2 / 720,
        (sine_theta + two_cosine_minus_two) / (theta2_nz * two_cosine_minus_two_nz),
    )

    b_tangent_vector = b.view(*size, 1) * tangent_vector
    jac = b_tangent_vector.view(*size, 3, 1) * tangent_vector.view(*size, 1, 3)

    half_ret = 0.5 * tangent_vector
    jac[..., 0, 1] -= half_ret[..., 2]
    jac[..., 1, 0] += half_ret[..., 2]
    jac[..., 0, 2] += half_ret[..., 1]
    jac[..., 2, 0] -= half_ret[..., 1]
    jac[..., 1, 2] -= half_ret[..., 0]
    jac[..., 2, 1] += half_ret[..., 0]

    diag_jac = torch.diagonal(jac, dim1=-1, dim2=-2)
    diag_jac += a.view(*size, 1)

    return jac, (b_tangent_vector,)


def _jlog_impl(group: torch.Tensor) -> Tuple[List[torch.Tensor], torch.Tensor]:
    check_group_tensor(group)
    tangent_vector, (theta, sine, cosine) = _log_impl_helper(group)
    jac, _ = _jlog_impl_helper(tangent_vector, theta, sine, cosine)
    return [jac], tangent_vector


class Log(lie_group.UnaryOperator):
    @classmethod
    def _forward_impl(cls, group):
        group: torch.Tensor = cast(torch.Tensor, group)
        tangent_vector = _log_impl(group)
        return tangent_vector

    @classmethod
    def setup_context(cls, ctx, inputs, outputs):
        # inputs is (group, ). outputs is tangent_vector
        ctx.save_for_backward(outputs, inputs[0])

    @classmethod
    def backward(cls, ctx, grad_output):
        group: torch.Tensor = ctx.saved_tensors[1]
        jacobians = 0.5 * _jlog_impl(group)[0][0]
        temp: torch.Tensor = _lift_autograd_fn(
            (jacobians.transpose(-2, -1) @ grad_output.unsqueeze(-1)).squeeze(-1)
        )
        return group @ temp


# TODO: Implement analytic backward for _jlog_impl
_log_autograd_fn = Log.apply
_jlog_autograd_fn = _jlog_impl


# -----------------------------------------------------------------------------
# Adjoint Transformation
# -----------------------------------------------------------------------------
def _adjoint_impl(group: torch.Tensor) -> torch.Tensor:
    check_group_tensor(group)
    return group.clone()


# NOTE: No jacobian is defined for the adjoint transformation
_jadjoint_impl = None


class Adjoint(lie_group.UnaryOperator):
    @classmethod
    def _forward_impl(cls, group):
        group: torch.Tensor = cast(torch.Tensor, group)
        return _adjoint_impl(group)

    @classmethod
    def backward(cls, ctx, grad_output):
        return grad_output


_adjoint_autograd_fn = Adjoint.apply
_jadjoint_autograd_fn = None


# -----------------------------------------------------------------------------
# Inverse
# -----------------------------------------------------------------------------
def _inverse_impl(group: torch.Tensor) -> torch.Tensor:
    check_group_tensor(group)
    return group.transpose(-1, -2)


_jinverse_impl = lie_group.JInverseImplFactory(_module)


class Inverse(lie_group.UnaryOperator):
    @classmethod
    def _forward_impl(cls, group):
        group: torch.Tensor = cast(torch.Tensor, group)
        return _inverse_impl(group)

    @classmethod
    def backward(cls, ctx, grad_output):
        return grad_output.transpose(-1, -2)


_inverse_autograd_fn = Inverse.apply
_jinverse_autograd_fn = _jinverse_impl


# -----------------------------------------------------------------------------
# Hat
# -----------------------------------------------------------------------------
def _hat_impl(tangent_vector: torch.Tensor) -> torch.Tensor:
    check_tangent_vector(tangent_vector)
    tangent_vector = tangent_vector.squeeze(-1)
    size = get_tangent_vector_size(tangent_vector)
    tensor = tangent_vector.new_zeros(*size, 3, 3)
    tensor[..., 0, 1] = -tangent_vector[..., 2]
    tensor[..., 0, 2] = tangent_vector[..., 1]
    tensor[..., 1, 2] = -tangent_vector[..., 0]
    tensor[..., 1, 0] = tangent_vector[..., 2]
    tensor[..., 2, 0] = -tangent_vector[..., 1]
    tensor[..., 2, 1] = tangent_vector[..., 0]

    return tensor


# NOTE: No jacobian is defined for the hat operator
_jhat_impl = None


class Hat(lie_group.UnaryOperator):
    @classmethod
    def _forward_impl(cls, tangent_vector):
        tangent_vector: torch.Tensor = cast(torch.Tensor, tangent_vector)
        ret = _hat_impl(tangent_vector)
        return ret

    @classmethod
    def backward(cls, ctx, grad_output):
        grad_output: torch.Tensor = cast(torch.Tensor, grad_output)
        return torch.stack(
            (
                grad_output[..., 2, 1] - grad_output[..., 1, 2],
                grad_output[..., 0, 2] - grad_output[..., 2, 0],
                grad_output[..., 1, 0] - grad_output[..., 0, 1],
            ),
            dim=-1,
        )


_hat_autograd_fn = Hat.apply
_jhat_autograd_fn = None


# -----------------------------------------------------------------------------
# Vee
# -----------------------------------------------------------------------------
def _vee_impl(tensor: torch.Tensor) -> torch.Tensor:
    check_hat_tensor(tensor)
    return 0.5 * torch.stack(
        (
            tensor[..., 2, 1] - tensor[..., 1, 2],
            tensor[..., 0, 2] - tensor[..., 2, 0],
            tensor[..., 1, 0] - tensor[..., 0, 1],
        ),
        dim=-1,
    )


# NOTE: No jacobian is defined for the vee operator
_jvee_impl = None


class Vee(lie_group.UnaryOperator):
    @classmethod
    def _forward_impl(cls, tangent_vector):
        tangent_vector: torch.Tensor = cast(torch.Tensor, tangent_vector)
        ret = _vee_impl(tangent_vector)
        return ret

    @classmethod
    def backward(cls, ctx, grad_output):
        grad_output: torch.Tensor = cast(torch.Tensor, grad_output)
        return 0.5 * _hat_autograd_fn(grad_output)


_vee_autograd_fn = Vee.apply
_jvee_autograd_fn = None


# -----------------------------------------------------------------------------
# Compose
# -----------------------------------------------------------------------------
def _compose_impl(group0: torch.Tensor, group1: torch.Tensor) -> torch.Tensor:
    check_group_tensor(group0)
    check_group_tensor(group1)
    return group0 @ group1


def _jcompose_impl(
    group0: torch.Tensor, group1: torch.Tensor
) -> Tuple[List[torch.Tensor], torch.Tensor]:
    check_group_tensor(group0)
    check_group_tensor(group1)
    ret = group0 @ group1
    size = get_group_size(ret)
    jac0 = group1.transpose(-1, -2).expand(*size, 3, 3).clone()
    jac1 = group0.new_zeros(*size, 3, 3)
    jac1[..., _DIAG_3_IDX, _DIAG_3_IDX] = 1
    return [jac0, jac1], ret


class Compose(lie_group.BinaryOperator):
    @classmethod
    def _forward_impl(cls, group0, group1):
        group0: torch.Tensor = cast(torch.Tensor, group0)
        group1: torch.Tensor = cast(torch.Tensor, group1)
        ret = _compose_impl(group0, group1)
        return ret

    @classmethod
    def setup_context(cls, ctx, inputs, outputs):
        # inputs is (group0, group1)
        ctx.save_for_backward(inputs[0], inputs[1])

    @classmethod
    def backward(cls, ctx, grad_output):
        group0, group1 = ctx.saved_tensors
        return (
            grad_output @ group1.transpose(-1, -2),
            group0.transpose(-1, -2) @ grad_output,
        )


_compose_autograd_fn = Compose.apply
_jcompose_autograd_fn = _jcompose_impl


# -----------------------------------------------------------------------------
# Transform From
# -----------------------------------------------------------------------------
def _transform_from_impl(group: torch.Tensor, tensor: torch.Tensor) -> torch.Tensor:
    check_group_tensor(group)
    check_transform_tensor(tensor)
    ret = group @ tensor.unsqueeze(-1)
    return ret.squeeze(-1)


def _jtransform_from_impl(
    group: torch.Tensor, tensor: torch.Tensor
) -> Tuple[List[torch.Tensor], torch.Tensor]:
    check_group_tensor(group)
    check_transform_tensor(tensor)
    ret = _transform_from_impl(group, tensor)
    size = get_transform_tensor_size(ret)
    jacobian_g = -group @ _hat_autograd_fn(tensor)
    jacobian_p = group
    jacobian_g = jacobian_g.expand(*size, 3, 3).clone()
    jacobian_p = jacobian_p.expand(*size, 3, 3).clone()
    return [jacobian_g, jacobian_p], ret


class TransformFrom(lie_group.BinaryOperator):
    @classmethod
    def _forward_impl(cls, group, tensor):
        group: torch.Tensor = cast(torch.Tensor, group)
        tensor: torch.Tensor = cast(torch.Tensor, tensor)
        ret = _transform_from_impl(group, tensor)
        return ret

    @classmethod
    def setup_context(cls, ctx, inputs, outputs):
        # inputs is (group, tensor)
        ctx.save_for_backward(inputs[0], inputs[1])

    @classmethod
    def backward(cls, ctx, grad_output):
        group: torch.Tensor = ctx.saved_tensors[0]
        tensor: torch.Tensor = ctx.saved_tensors[1]
        grad_output: torch.Tensor = grad_output.unsqueeze(-1)
        tensor_size = get_transform_tensor_size(tensor)
        grad_input0 = grad_output @ tensor.view(*tensor_size, 1, 3)
        grad_input1 = group[..., :3].transpose(-1, -2) @ grad_output
        return grad_input0, grad_input1.squeeze(-1)


_transform_from_autograd_fn = TransformFrom.apply
_jtransform_from_autograd_fn = _jtransform_from_impl


# -----------------------------------------------------------------------------
# Unit Quaternion to Rotation Matrix
# -----------------------------------------------------------------------------
def _quaternion_to_rotation_impl(quaternion: torch.Tensor) -> torch.Tensor:
    check_unit_quaternion(quaternion)

    quaternion = quaternion / torch.norm(quaternion, dim=-1, keepdim=True)
    w = quaternion[..., 0]
    x = quaternion[..., 1]
    y = quaternion[..., 2]
    z = quaternion[..., 3]

    q00 = w * w
    q01 = w * x
    q02 = w * y
    q03 = w * z
    q11 = x * x
    q12 = x * y
    q13 = x * z
    q22 = y * y
    q23 = y * z
    q33 = z * z

    size = quaternion.shape[:-1]
    ret = quaternion.new_zeros(*size, 3, 3)
    ret[..., 0, 0] = q00 + q11 - q22 - q33
    ret[..., 0, 1] = 2 * (q12 - q03)
    ret[..., 0, 2] = 2 * (q13 + q02)
    ret[..., 1, 0] = 2 * (q12 + q03)
    ret[..., 1, 1] = q00 - q11 + q22 - q33
    ret[..., 1, 2] = 2 * (q23 - q01)
    ret[..., 2, 0] = 2 * (q13 - q02)
    ret[..., 2, 1] = 2 * (q23 + q01)
    ret[..., 2, 2] = q00 - q11 - q22 + q33
    return ret


def _jquaternion_to_rotation_impl(
    quaternion: torch.Tensor,
) -> Tuple[List[torch.Tensor], torch.Tensor]:
    if quaternion.ndim == 1:
        quaternion = quaternion.unsqueeze(0)
    check_unit_quaternion(quaternion)

    quaternion_norm = torch.norm(quaternion, dim=-1, keepdim=True)
    quaternion = quaternion / quaternion_norm
    w = quaternion[..., 0]
    x = quaternion[..., 1]
    y = quaternion[..., 2]
    z = quaternion[..., 3]

    q00 = w * w
    q01 = w * x
    q02 = w * y
    q03 = w * z
    q11 = x * x
    q12 = x * y
    q13 = x * z
    q22 = y * y
    q23 = y * z
    q33 = z * z

    size = quaternion.shape[:-1]
    ret = quaternion.new_zeros(*size, 3, 3)
    ret[..., 0, 0] = q00 + q11 - q22 - q33
    ret[..., 0, 1] = 2 * (q12 - q03)
    ret[..., 0, 2] = 2 * (q13 + q02)
    ret[..., 1, 0] = 2 * (q12 + q03)
    ret[..., 1, 1] = q00 - q11 + q22 - q33
    ret[..., 1, 2] = 2 * (q23 - q01)
    ret[..., 2, 0] = 2 * (q13 - q02)
    ret[..., 2, 1] = 2 * (q23 + q01)
    ret[..., 2, 2] = q00 - q11 - q22 + q33

    temp = -2 * quaternion / quaternion_norm
    jac = quaternion.new_zeros(*size, 3, 4)
    jac[..., :, :1] = temp[..., 1:].view(*size, 3, 1)
    jac[..., :, 1:] = _hat_autograd_fn(temp[..., 1:])
    jac[..., 0, 1] = -temp[..., 0]
    jac[..., 1, 2] = -temp[..., 0]
    jac[..., 2, 3] = -temp[..., 0]

    return [jac], ret


class QuaternionToRotation(lie_group.UnaryOperator):
    @classmethod
    def _forward_impl(cls, quaternion):
        quaternion: torch.Tensor = cast(torch.Tensor, quaternion)
        ret = _quaternion_to_rotation_impl(quaternion)
        return ret

    @classmethod
    def setup_context(cls, ctx, inputs, outputs):
        # inputs is (quaternion,). outputs is rotation
        ctx.save_for_backward(inputs[0], outputs)

    @classmethod
    def backward(cls, ctx, grad_output):
        quaternion: torch.Tensor = ctx.saved_tensors[0]
        group: torch.Tensor = ctx.saved_tensors[1]
        jacs = _jquaternion_to_rotation_impl(quaternion)[0][0]
        dR = group.transpose(-1, -2) @ grad_output
        size = quaternion.shape[:-1]
        grad_input = jacs.transpose(-1, -2) @ torch.stack(
            (
                dR[..., 2, 1] - dR[..., 1, 2],
                dR[..., 0, 2] - dR[..., 2, 0],
                dR[..., 1, 0] - dR[..., 0, 1],
            ),
            dim=-1,
        ).view(*size, 3, 1)
        return grad_input.view_as(quaternion)


_quaternion_to_rotation_autograd_fn = QuaternionToRotation.apply
_jquaternion_to_rotation_autograd_fn = _jquaternion_to_rotation_impl


# -----------------------------------------------------------------------------
# Lift
# -----------------------------------------------------------------------------
def _lift_impl(tensor: torch.Tensor) -> torch.Tensor:
    check_lift_tensor(tensor)
    ret = tensor.new_zeros(tensor.shape[:-1] + (3, 3))
    ret[..., 0, 1] = -tensor[..., 2]
    ret[..., 0, 2] = tensor[..., 1]
    ret[..., 1, 2] = -tensor[..., 0]
    ret[..., 1, 0] = tensor[..., 2]
    ret[..., 2, 0] = -tensor[..., 1]
    ret[..., 2, 1] = tensor[..., 0]

    return ret


# NOTE: No jacobian is defined for the project operator
_jlift_impl = None


class Lift(lie_group.UnaryOperator):
    @classmethod
    def _forward_impl(cls, tensor):
        tensor: torch.Tensor = cast(torch.Tensor, tensor)
        ret = _lift_impl(tensor)
        return ret

    @classmethod
    def backward(cls, ctx, grad_output):
        grad_output: torch.Tensor = cast(torch.Tensor, grad_output)
        return _project_autograd_fn(grad_output)


_lift_autograd_fn = Lift.apply
_jlift_autograd_fn = None


# -----------------------------------------------------------------------------
# Project
# -----------------------------------------------------------------------------
def _project_impl(tensor: torch.Tensor) -> torch.Tensor:
    check_project_tensor(tensor)
    return torch.stack(
        (
            tensor[..., 2, 1] - tensor[..., 1, 2],
            tensor[..., 0, 2] - tensor[..., 2, 0],
            tensor[..., 1, 0] - tensor[..., 0, 1],
        ),
        dim=-1,
    )


# NOTE: No jacobian is defined for the project operator
_jproject_impl = None


class Project(lie_group.UnaryOperator):
    @classmethod
    def _forward_impl(cls, tensor):
        tensor: torch.Tensor = cast(torch.Tensor, tensor)
        ret = _project_impl(tensor)
        return ret

    @classmethod
    def backward(cls, ctx, grad_output):
        grad_output: torch.Tensor = cast(torch.Tensor, grad_output)
        return _lift_autograd_fn(grad_output)


_project_autograd_fn = Project.apply
_jproject_autograd_fn = None


# -----------------------------------------------------------------------------
# Left Act
# -----------------------------------------------------------------------------
def _left_act_impl(
    group: torch.Tensor, tensor: torch.Tensor, dim_out: Optional[int] = None
) -> torch.Tensor:
    check_group_tensor(group)
    check_left_act_tensor(tensor)

    dim_out = tensor.ndim - group.ndim if dim_out is None else dim_out

    if group.ndim + dim_out > tensor.ndim:
        tensor = fill_dims(tensor, group.ndim + dim_out)

    permuted_dim = permute_op_dim(tensor.ndim, dim_out, 2)
    unpermuted_dim = unpermute_op_dim(tensor.ndim, dim_out, 2)
    tensor = tensor.permute(permuted_dim)
    return (group @ tensor).permute(unpermuted_dim)


def _left_act_backward_helper(
    group: torch.Tensor, tensor: torch.Tensor, dim_out: int, grad_output: torch.Tensor
):
    if group.ndim + dim_out > tensor.ndim:
        tensor = fill_dims(tensor, group.ndim + dim_out)

    permuted_dim = permute_op_dim(tensor.ndim, dim_out, 2)
    unpermuted_dim = unpermute_op_dim(tensor.ndim, dim_out, 2)
    tensor = tensor.permute(permuted_dim)
    grad_output = grad_output.permute(permuted_dim)
    jac_group = (grad_output @ tensor.transpose(-1, -2)).permute(unpermuted_dim)
    jac_tensor = (group.transpose(-1, -2) @ grad_output).permute(unpermuted_dim)
    if dim_out > 0:
        dim = list(range(tensor.ndim - 2 - dim_out, tensor.ndim - 2))
        jac_group = jac_group.sum(dim)
    return jac_group, jac_tensor, None


class LeftAct(lie_group.GradientOperator):
    @classmethod
    def _forward_impl(cls, group, tensor, dim_out):
        group: torch.Tensor = cast(torch.Tensor, group)
        tensor: torch.Tensor = cast(torch.Tensor, tensor)
        ret = _left_act_impl(group, tensor, dim_out)
        return ret

    @classmethod
    def setup_context(cls, ctx, inputs, outputs):
        # inputs is (group, tensor)
        ctx.save_for_backward(inputs[0], inputs[1])
        ctx.dim_out = inputs[2]

    @classmethod
    def backward(cls, ctx, grad_output):
        group, tensor = ctx.saved_tensors
        dim_out = ctx.dim_out
        dim_out: int = tensor.ndim - group.ndim if dim_out is None else dim_out
        return _left_act_backward_helper(group, tensor, dim_out, grad_output)


def _left_act_autograd_fn(
    group: torch.Tensor, tensor: torch.Tensor, dim_out: Optional[int] = None
):
    return LeftAct.apply(group, tensor, dim_out)


_jleft_act_autograd_fn = None


# -----------------------------------------------------------------------------
# Left Project
# -----------------------------------------------------------------------------
_left_project_impl = lie_group.LeftProjectImplFactory(_module)
_jleft_project_impl = None


def _left_project_backward_helper(
    group: torch.Tensor,
    tensor: torch.Tensor,
    dim_out: int,
    grad_output_lifted: torch.Tensor,
):
    jac_group, jac_tensor, _ = _left_act_backward_helper(
        group.transpose(-1, -2), tensor, dim_out, grad_output_lifted
    )
    return jac_group.transpose(-1, -2), jac_tensor, None


class LeftProject(lie_group.GradientOperator):
    @classmethod
    def _forward_impl(cls, group, tensor, dim_out):
        group = cast(torch.Tensor, group)
        tensor = cast(torch.Tensor, tensor)
        ret = _left_project_impl(group, tensor, dim_out)
        return ret

    @classmethod
    def setup_context(cls, ctx, inputs, outputs):
        # inputs is (group, tensor)
        ctx.save_for_backward(inputs[0], inputs[1])
        ctx.dim_out = inputs[2]

    @classmethod
    def backward(cls, ctx, grad_output):
        group, tensor = ctx.saved_tensors
        dim_out = ctx.dim_out
        dim_out: int = tensor.ndim - group.ndim if dim_out is None else dim_out
        grad_output_lifted = _lift_autograd_fn(grad_output)
        return _left_project_backward_helper(group, tensor, dim_out, grad_output_lifted)


def _left_project_autograd_fn(
    group: torch.Tensor, tensor: torch.Tensor, dim_out: Optional[int] = None
):
    return LeftProject.apply(group, tensor, dim_out)


_jleft_project_autograd_fn = _jleft_project_impl


# -----------------------------------------------------------------------------
# Normalize
# -----------------------------------------------------------------------------
def _normalize_impl_helper(tensor: torch.Tensor):
    check_group_shape(tensor)
    size = tensor.shape[:-2]
    u, s, v = torch.svd(tensor)
    sign = torch.det(u @ v).view(*size, 1, 1)
    vt = torch.cat(
        (v[..., :2], torch.where(sign > 0, v[..., 2:], -v[..., 2:])), dim=-1
    ).transpose(-1, -2)
    return u @ vt, {"u": u, "s": s, "v": v, "sign": sign}


def _normalize_impl(tensor: torch.Tensor) -> torch.Tensor:
    return _normalize_impl_helper(tensor)[0]


def _normalize_backward_helper(
    u: torch.Tensor,
    s: torch.Tensor,
    v: torch.Tensor,
    sign: torch.Tensor,
    grad_output: torch.Tensor,
) -> torch.Tensor:
    def _skew_symm(tensor: torch.Tensor) -> torch.Tensor:
        return tensor - tensor.transpose(-1, -2)

    size = u.shape[:-2]
    ut = u.transpose(-1, -2)
    vt = v.transpose(-1, -2)
    grad_u: torch.Tensor = grad_output @ torch.cat(
        (v[..., :2], v[..., 2:] @ sign), dim=-1
    )
    grad_v: torch.Tensor = grad_output.transpose(-1, -2) @ torch.cat(
        (u[..., :2], u[..., 2:] @ sign), dim=-1
    )
    s_squared: torch.Tensor = s.pow(2)
    F = s_squared.view(*size, 1, 3).expand(*size, 3, 3) - s_squared.view(
        *size, 3, 1
    ).expand(*size, 3, 3)
    F = torch.where(F == 0, constants._INF, F)
    F = F.pow(-1)

    u_term: torch.Tensor = u @ (F * _skew_symm(ut @ grad_u))
    # The next 3 lines are equivalent to u_term = torch.einsum("n...ij, n...j->n...ij", u_term, s).
    # This implementation is compatible for vectorize=True in torch.autograd.functional.jacobian.
    permuted_u_term_dim = permute_op_dim(u_term.dim(), 1, 1)
    unpermuted_u_term_dim = unpermute_op_dim(u_term.dim(), 1, 1)
    u_term = (u_term.permute(permuted_u_term_dim) * s).permute(unpermuted_u_term_dim)
    u_term = u_term @ vt

    v_term: torch.Tensor = (F * _skew_symm(vt @ grad_v)) @ vt
    # The next 3 lines are equivalent to v_term = torch.einsum("n...i, n...ij->n...ij", s, v_term).
    # This implementation is compatible for vectorize=True in torch.autograd.functional.jacobian.
    permuted_v_term_dim = permute_op_dim(v_term.dim(), 1, 0)
    unpermuted_v_term_dim = unpermute_op_dim(v_term.dim(), 1, 0)
    v_term = (s * v_term.permute(permuted_v_term_dim)).permute(unpermuted_v_term_dim)
    v_term = u @ v_term

    return u_term + v_term


class Normalize(lie_group.UnaryOperator):
    @classmethod
    def _forward_impl(cls, tensor):
        tensor: torch.Tensor = tensor
        output, svd_info = _normalize_impl_helper(tensor)
        return output, svd_info

    @staticmethod
    def setup_context(ctx, inputs, outputs):
        # outputs is (normalized_out, svd_info)
        svd_info = outputs[1]
        ctx.save_for_backward(
            svd_info["u"], svd_info["s"], svd_info["v"], svd_info["sign"]
        )

    @classmethod
    def backward(cls, ctx, grad_output, _):
        u, s, v, sign = ctx.saved_tensors
        return _normalize_backward_helper(u, s, v, sign, grad_output), None


def _normalize_autograd_fn(tensor: torch.Tensor):
    return Normalize.apply(tensor)[0]


_jnormalize_autograd_fn = None


_fns = lie_group.LieGroupFns(_module)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
import threading
import warnings
from typing import Any, Callable, List, Optional, Tuple
from typing import cast as type_cast

import torch
from torch.utils._pytree import tree_flatten, tree_map_only
from torch.types import Number

from .functional.constants import DeviceType
from .functional.lie_group import UnaryOperatorOpFnType
from .types import (
    ltype as _ltype,
    Device,
    SE3,
    SO3,
    TensorType,
    _IdentityFnType,
    _JFnReturnType,
    _RandFnType,
)


class _LieAsEuclideanContext:
    contexts = threading.local()

    @classmethod
    def get_context(cls):
        if not hasattr(cls.contexts, "as_euclidean"):
            cls.contexts.as_euclidean = False
        return cls.contexts.as_euclidean

    @classmethod
    def set_context(cls, as_euclidean: bool):
        cls.contexts.as_euclidean = as_euclidean


class as_euclidean:
    def __init__(self) -> None:
        self.prev = _LieAsEuclideanContext.get_context()
        _LieAsEuclideanContext.set_context(True)

    def __enter__(self) -> None:
        pass

    def __exit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None:
        _LieAsEuclideanContext.set_context(self.prev)


def euclidean_enabled() -> bool:
    return _LieAsEuclideanContext.get_context()


class _LieTensorBase(torch.Tensor):
    def __new__(
        cls,
        data: torch.Tensor,
        ltype: _ltype,
        requires_grad: bool = False,
        _shared_memory: bool = False,
    ):
        return cls._make_subclass(
            cls, data if _shared_memory else data.clone(), requires_grad  # type: ignore
        )

    @property
    def _t(self) -> torch.Tensor:
        return self.as_subclass(torch.Tensor)

    def __init__(
        self,
        data: Any,
        ltype: _ltype,
        requires_grad: Optional[bool] = None,
        _shared_memory: bool = False,
    ):
        self.ltype = ltype
        self.ltype._fn_lib.check_group_tensor(self.as_subclass(torch.Tensor))

    def __repr__(self) -> str:  # type: ignore
        return f"LieTensor({self._t}, ltype=lie.{self.ltype})"

    @classmethod
    def to_torch(cls, args: Any):
        return tree_map_only(cls, lambda x: x._t, args)

    @staticmethod
    def resolve_ltype(t: Any):
        return t.ltype if hasattr(t, "ltype") else None

    @classmethod
    def get_ltype(cls, args):
        ltypes = [cls.resolve_ltype(a) for a in tree_flatten(args)[0]]
        ltypes = set(x for x in ltypes if x is not None)
        if len(ltypes) > 1:
            raise ValueError(
                f"All LieTensors must be of the same ltype. " f"But found {ltypes}"
            )
        return next(iter(ltypes)) if len(ltypes) > 0 else None

    @classmethod
    def _torch_func_impl_eucl(cls, func, types, args=(), kwargs=None, raw_tensor=False):
        kwargs = kwargs or {}
        torch_args = cls.to_torch(args)
        torch_kwargs = cls.to_torch(kwargs)
        ltype = cls.get_ltype(args)
        ret = func(*torch_args, **torch_kwargs)
        if raw_tensor:
            return ret
        return tree_map_only(torch.Tensor, lambda x: cls(x, ltype), ret)


_EUCLID_GRAD_MARKER = "_lie_euclidean_grad"


# Lightweight wrapper so that LieTensor.add_ knows that this is a gradient
# that should be left projected and retracted
class _EuclideanGrad(torch.Tensor):
    _lie_euclidean_grad = True

    @classmethod
    def __torch_function__(cls, func, types, args=(), kwargs=None):
        # Need to do this marking thing because torch optimizers use
        # in-place operations, so in many cases the descent direction won't be
        # an object of type _EuclideanGrad. With this, anything that touches
        # an _EuclideanGrad becomes marked and can be used to update LieTensors.
        def _mark(x):
            if isinstance(x, torch.Tensor):
                setattr(x, _EUCLID_GRAD_MARKER, True)

        for a in tree_flatten(args)[0]:
            _mark(a)

        return super().__torch_function__(func, types, args, kwargs or {})


def _LIE_TENSOR_GRAD_ERROR_MSG(func):
    return (
        f"LieTensor.{func.__name__} is only supported for "
        "tensors resulting from euclidean gradient computations."
    )


class LieTensor(_LieTensorBase):
    # These are operations where calling super().__torch_function__() on the
    # LieTensor itself is safe
    _SAFE_SUPER_OPS: List[Callable] = [
        torch.Tensor.device.__get__,  # type: ignore
        torch.Tensor.dtype.__get__,  # type: ignore
        torch.Tensor.is_leaf.__get__,  # type: ignore
        torch.Tensor.is_mps.__get__,  # type: ignore
        torch.Tensor.is_sparse.__get__,  # type: ignore
        torch.Tensor.is_quantized.__get__,  # type: ignore
        torch.Tensor.grad.__get__,  # type: ignore
        torch.Tensor.grad.__set__,  # type: ignore
        torch.Tensor.layout.__get__,  # type: ignore
        torch.Tensor.requires_grad.__get__,  # type: ignore
        torch.Tensor.retains_grad.__get__,  # type: ignore
        torch.Tensor.storage,  # type: ignore
        torch.Tensor.__getitem__,  # type: ignore
        torch.Tensor.clone,
        torch.Tensor.__format__,
        torch.Tensor.new_tensor,
        torch.Tensor.to,
        torch.cat,
        torch.is_complex,
        # torch.stack  # requires arbitrary batch support
    ]

    # These are operations where calling super().__torch_function__() on the
    # lie_tensor._t view is safe, and that require a return value of
    # raw torch.Tensor, because their return values don't make sense for LieTensors.
    _SAFE_AS_EUCL_OPS: List[Callable] = [
        torch.zeros_like,
        torch.ones_like,
        torch.Tensor.new_zeros,
        torch.Tensor.new_ones,
        torch.allclose,
        torch.isclose,
        torch.Tensor.grad_fn.__get__,  # type: ignore
        torch.Tensor.shape.__get__,  # type: ignore
        torch.Tensor.ndim.__get__,  # type: ignore
    ]

    def __init__(
        self,
        tensor: torch.Tensor,
        ltype: _ltype,
        requires_grad: Optional[bool] = None,
        _shared_memory: bool = False,
    ):
        super().__init__(
            tensor,
            ltype,
            requires_grad=requires_grad,
            _shared_memory=_shared_memory,
        )

    @staticmethod
    def from_tensor(tensor: torch.Tensor, ltype: _ltype) -> "LieTensor":
        return _FromTensor.apply(tensor, ltype)

    @classmethod
    def _torch_function_impl_lie(cls, func, types, args=(), kwargs=None):
        if func in LieTensor._SAFE_AS_EUCL_OPS:
            return super()._torch_func_impl_eucl(
                func, types, args, kwargs, raw_tensor=True
            )
        if func in LieTensor._SAFE_SUPER_OPS:
            ltype = cls.get_ltype(args)
            ret = super().__torch_function__(func, types, args, kwargs or {})
            if func == torch.Tensor.grad.__get__:
                return _EuclideanGrad(ret) if ret is not None else ret
            # tree-map to set the ltypes correctly
            if func == torch.Tensor.__getitem__:
                if isinstance(args[1], tuple):
                    raise NotImplementedError(
                        "LieTensor currently only supports slicing the batch dimension."
                    )
                assert ret.ndim in [2, 3]
                if ret.ndim == 2:
                    ret = ret._t.unsqueeze(0)
            return tree_map_only(torch.Tensor, lambda x: from_tensor(x, ltype), ret)
        raise NotImplementedError(
            "Tried to call a torch function not supported by LieTensor. "
            "If trying to operate on the raw tensor data, please use group._t, "
            "or run inside the context lie.as_euclidean()."
        )

    @classmethod
    def __torch_function__(cls, func, types, args=(), kwargs=None):
        if euclidean_enabled():
            return super()._torch_func_impl_eucl(
                func, types, args, kwargs, raw_tensor=True
            )
        else:
            return cls._torch_function_impl_lie(func, types, args, kwargs)

    def _check_ltype(self, other: "LieTensor", op_name: str):
        if other.ltype != self.ltype:
            raise ValueError("f{op_name} requires both tensors to have same ltype.")

    # Returns a new LieTensor with the given data and the same ltype as self
    def new_lietensor(  # type: ignore
        self, args: Any, device: Device = None, dtype: Optional[torch.dtype] = None
    ) -> "LieTensor":
        if isinstance(args, LieTensor):
            warnings.warn(
                "To copy construct from a LieTensor, it is recommended to use "
                "lie_tensor.clone().detach() instead of lie_tensor.new_lietensor(t)",
                UserWarning,
            )
            tensor_arg = args._t
            dtype = dtype or self.dtype
        elif isinstance(args, torch.Tensor):
            tensor_arg = args
            dtype = dtype or args.dtype
        else:
            tensor_arg = torch.as_tensor(args)
            dtype = dtype or torch.get_default_dtype()
        return LieTensor(tensor_arg.to(device=device, dtype=dtype), ltype=self.ltype)

    # ------------------------------------------------------
    # ------ Operators
    # ------------------------------------------------------
    # For the next ones the output also depends on self's data
    def log(self) -> torch.Tensor:
        return self.ltype._fn_lib.log(self._t)

    def adj(self) -> torch.Tensor:
        return self.ltype._fn_lib.adj(self._t)

    def inv(self) -> "LieTensor":
        return self.from_tensor(self.ltype._fn_lib.inv(self._t), self.ltype)

    def compose(self, other: "LieTensor") -> "LieTensor":
        self._check_ltype(other, "compose")
        return self.from_tensor(
            self.ltype._fn_lib.compose(self._t, other._t), self.ltype
        )

    def transform_from(self, point: torch.Tensor) -> torch.Tensor:
        return self.ltype._fn_lib.transform_from(self._t, point)

    def left_act(self, matrix: torch.Tensor) -> torch.Tensor:
        return self.ltype._fn_lib.left_act(self._t, matrix)

    def left_project(self, matrix: torch.Tensor) -> torch.Tensor:
        return self.ltype._fn_lib.left_project(self._t, matrix)

    # ------------------------------------------------------
    # Operator Jacobians
    # ------------------------------------------------------
    def _unary_jop_base(
        self,
        input0: torch.Tensor,
        fn: UnaryOperatorOpFnType,
        out_is_group: bool = True,
    ) -> _JFnReturnType:
        jacs: List[torch.Tensor] = []
        op_res: TensorType = fn(input0, jacobians=jacs)
        if out_is_group:
            op_res = self.from_tensor(op_res, self.ltype)
        return jacs, op_res

    def jlog(self) -> _JFnReturnType:
        return self._unary_jop_base(self._t, self.ltype._fn_lib.log, out_is_group=False)

    def jinv(self) -> _JFnReturnType:
        return self._unary_jop_base(self._t, self.ltype._fn_lib.inv)

    def jcompose(self, other: "LieTensor") -> _JFnReturnType:
        self._check_ltype(other, "jcompose")
        jacs: List[torch.Tensor] = []
        op_res = self.from_tensor(
            self.ltype._fn_lib.compose(self._t, other._t, jacobians=jacs), self.ltype
        )
        return jacs, op_res

    def jtransform_from(self, point: torch.Tensor) -> _JFnReturnType:
        jacs: List[torch.Tensor] = []
        op_res = self.ltype._fn_lib.transform_from(self._t, point, jacobians=jacs)
        return jacs, op_res

    def _no_jop(self, input0: TensorType) -> _JFnReturnType:
        raise NotImplementedError

    jadjoint = _no_jop
    jhat = _no_jop
    jvee = _no_jop
    jlift = _no_jop
    jproject = _no_jop
    jleft_act = _no_jop
    jleft_project = _no_jop

    def retract(self, delta: TensorType) -> "LieTensor":
        if not isinstance(delta, torch.Tensor):
            raise TypeError(
                "LieTensor.retract() expects a single torch.Tensor argument."
            )
        return self.compose(self.ltype.exp(delta))

    def local(self, other: "LieTensor") -> torch.Tensor:
        if not isinstance(other, LieTensor):
            raise TypeError(
                f"Incorrect argument for LieTensor.local(). "
                f"Expected LieTensor, but got {type(other)}."
            )
        if not other.ltype == self.ltype:
            raise ValueError(
                f"Incorrect ltype for local. Expected {self.ltype}, "
                f"received {other.ltype}."
            )
        return self.inv().compose(other).log()

    # ------------------------------------------------------
    # Overloaded python and torch operators
    # ------------------------------------------------------
    def __mul__(self, other: TensorType) -> "LieTensor":
        if not isinstance(other, LieTensor):
            raise TypeError(
                f"Incorrect argument for '*' operator. "
                f"Expected LieTensor, but got {type(other)}"
            )
        return type_cast(LieTensor, other).compose(self)

    def __matmul__(self, point: TensorType) -> torch.Tensor:
        if isinstance(point, LieTensor):
            raise TypeError(
                "Incorrect argument for '@' operator. "
                "Expected a torch.Tensor (x, y, z), but got a LieTensor."
            )
        return self.transform_from(point)

    def set_(self, tensor: "LieTensor"):  # type: ignore
        if not isinstance(tensor, LieTensor):
            raise RuntimeError(
                "LieTensor.set_ is only supported for LieTensor arguments."
            )
        if not tensor.ltype == self.ltype:
            raise RuntimeError(
                f"Tried to set a tensor of type {self.ltype} with a "
                f"tensor of type {tensor.ltype}"
            )
        super().set_(tensor)  # type: ignore

    def add_(self, tensor: torch.Tensor, *, alpha: Number = 1.0):  # type: ignore
        if hasattr(tensor, _EUCLID_GRAD_MARKER):
            grad = self.ltype._fn_lib.left_project(self._t, tensor)
            res = self.retract(grad * alpha)
            self.set_(res)
        else:
            raise RuntimeError(_LIE_TENSOR_GRAD_ERROR_MSG(self.add_))

    def addcdiv_(
        self, tensor1: torch.Tensor, tensor2: torch.Tensor, value: Number = 1.0
    ):
        can_do = False
        for t in [tensor1, tensor2]:
            if hasattr(t, _EUCLID_GRAD_MARKER):
                can_do = True
        if can_do:
            self.add_(_EuclideanGrad(value * tensor1 / tensor2))
        else:
            raise RuntimeError(_LIE_TENSOR_GRAD_ERROR_MSG(self.addcdiv_))

    def addcmul_(
        self, tensor1: torch.Tensor, tensor2: torch.Tensor, value: Number = 1.0
    ):
        can_do = False
        for t in [tensor1, tensor2]:
            if hasattr(t, _EUCLID_GRAD_MARKER):
                can_do = True
        if can_do:
            self.add_(_EuclideanGrad(value * tensor1 * tensor2))
        else:
            raise RuntimeError(_LIE_TENSOR_GRAD_ERROR_MSG(self.addcmul_))


# ----------------------------
# Tensor creation functions
# ----------------------------
def as_lietensor(
    data: Any,
    ltype: Optional[_ltype] = None,
    dtype: torch.dtype = torch.float,
    device: DeviceType = None,
) -> _LieTensorBase:
    if isinstance(data, LieTensor):
        if data.dtype == dtype and data.device == torch.device(device or "cpu"):
            return data
        return type_cast(LieTensor, data.to(device=device, dtype=dtype))
    if ltype is None:
        raise ValueError("ltype must be provided.")
    return type_cast(
        LieTensor,
        LieTensor.from_tensor(data, ltype=ltype).to(device=device, dtype=dtype),
    )


def cast(
    data: Any,
    ltype: Optional[_ltype] = None,
    dtype: torch.dtype = torch.float,
    device: DeviceType = None,
) -> _LieTensorBase:
    return as_lietensor(data, ltype=ltype, dtype=dtype, device=device)


from_tensor: Callable[[torch.Tensor, _ltype], LieTensor] = LieTensor.from_tensor


class _FromTensor(torch.autograd.Function):
    @staticmethod
    def forward(ctx: Any, tensor: torch.Tensor, ltype: _ltype) -> LieTensor:  # type: ignore
        return LieTensor(
            tensor, ltype, requires_grad=tensor.requires_grad, _shared_memory=True
        )

    @staticmethod
    def backward(ctx: Any, grad_output: torch.Tensor) -> Tuple[torch.Tensor, None]:  # type: ignore
        return grad_output, None


def _build_random_fn(op_name: str, ltype: _ltype) -> _RandFnType:
    assert op_name in ["rand", "randn"]

    def fn(
        *size: Any,
        generator: Optional[torch.Generator] = None,
        dtype: Optional[torch.dtype] = None,
        device: DeviceType = None,
        requires_grad: bool = False,
    ) -> LieTensor:
        good = all([isinstance(a, int) for a in size])
        if not good:
            arg_types = " ".join(type(a).__name__ for a in size)
            raise TypeError(
                f"{op_name}() received invalid combination of arguments - "
                f"got ({arg_types}), but expected (tuple of ints)."
            )
        fn = getattr(ltype._fn_lib, op_name)
        return LieTensor(
            fn(
                *size,
                generator=generator,
                dtype=dtype,
                device=device,
                requires_grad=requires_grad,
            ),
            ltype,
            requires_grad=requires_grad,
            _shared_memory=True,
        )

    return fn


def _build_identity_fn(ltype: _ltype) -> _IdentityFnType:
    def fn(
        *size: Any,
        dtype: Optional[torch.dtype] = None,
        device: DeviceType = None,
        requires_grad: bool = False,
    ) -> LieTensor:
        good = all([isinstance(a, int) for a in size])
        if not good:
            arg_types = " ".join(type(a).__name__ for a in size)
            raise TypeError(
                f"identity() received invalid combination of arguments - "
                f"got ({arg_types}), but expected (tuple of ints)."
            )
        return LieTensor(
            ltype._fn_lib.identity(
                *size,
                dtype=dtype,
                device=device,
                requires_grad=requires_grad,
            ),
            ltype,
            requires_grad=requires_grad,
            _shared_memory=True,
        )

    return fn


def _build_call_impl(ltype: _ltype) -> Callable[[torch.Tensor], LieTensor]:
    def fn(tensor: torch.Tensor) -> LieTensor:
        return LieTensor(tensor, ltype)

    return fn


SE3.rand = _build_random_fn("rand", SE3)
SE3.randn = _build_random_fn("randn", SE3)
SE3.identity = _build_identity_fn(SE3)
SE3._call_impl = _build_call_impl(SE3)
SO3.rand = _build_random_fn("rand", SO3)
SO3.randn = _build_random_fn("randn", SO3)
SO3.identity = _build_identity_fn(SO3)
SO3._call_impl = _build_call_impl(SO3)
SE3._create_lie_tensor = SO3._create_lie_tensor = LieTensor


def log(group: LieTensor) -> torch.Tensor:
    return group.log()


def adj(group: LieTensor) -> torch.Tensor:
    return group.adj()


def inv(group: LieTensor) -> LieTensor:
    return group.inv()


def compose(group1: LieTensor, group2: LieTensor) -> LieTensor:
    return group1.compose(group2)


def transform_from(group1: LieTensor, tensor: torch.Tensor) -> torch.Tensor:
    return group1.transform_from(tensor)


def left_act(group: LieTensor, matrix: torch.Tensor) -> torch.Tensor:
    return group.left_act(matrix)


def left_project(group: LieTensor, matrix: torch.Tensor) -> torch.Tensor:
    return group.left_project(matrix)


def jlog(group: LieTensor) -> _JFnReturnType:
    return group.jlog()


def jinv(group: LieTensor) -> _JFnReturnType:
    return group.jinv()


def jcompose(group1: LieTensor, group2: LieTensor) -> _JFnReturnType:
    return group1.jcompose(group2)


def jtransform_from(group1: LieTensor, tensor: torch.Tensor) -> _JFnReturnType:
    return group1.jtransform_from(tensor)


def retract(group: LieTensor, delta: TensorType) -> LieTensor:
    return group.retract(delta)


def local(group1: LieTensor, group2: LieTensor) -> torch.Tensor:
    return group1.local(group2)



#!/usr/bin/env python
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
import setuptools
from pathlib import Path

lie_path = Path("torchlie")
with open(lie_path / "__init__.py", "r") as f:
    for line in f:
        if "__version__" in line:
            version = line.split("__version__ = ")[1].rstrip().strip('"')

with open("README.md", "r") as fh:
    long_description = fh.read()


setuptools.setup(
    name="torchlie",
    version=version,
    author="Meta Research",
    description="Torch extension for differentiable Lie groups.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/facebookresearch/theseus/lie",
    keywords="lie groups, differentiable optimization",
    packages=["torchlie", "torchlie.functional"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
    python_requires=">=3.8",
)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
__version__ = "0.0.1.rc1"

from .types import ltype, SE3, SO3
from .lie_tensor import (  # usort: skip
    LieTensor,
    adj,
    as_euclidean,
    as_lietensor,
    cast,
    compose,
    from_tensor,
    inv,
    jcompose,
    jinv,
    jlog,
    jtransform_from,
    left_act,
    left_project,
    local,
    log,
    transform_from,
)
from .options import reset_global_options, set_global_options



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
from typing import Any, Dict
from dataclasses import dataclass

import torch


def _CHECK_DTYPE_SUPPORTED(dtype):
    if dtype not in [torch.float32, torch.float64]:
        raise ValueError(
            f"Unsupported data type {dtype}. "
            "Theseus only supports 32- and 64-bit tensors."
        )


@dataclass
class _TorchLieOptions:
    so3_near_pi_eps_float32: float = 0
    so3_near_zero_eps_float32: float = 0
    so3_matrix_eps_float32: float = 0
    so3_quat_eps_float32: float = 0
    so3_hat_eps_float32: float = 0
    se3_hat_eps_float32: float = 0
    so3_near_pi_eps_float64: float = 0
    so3_near_zero_eps_float64: float = 0
    so3_matrix_eps_float64: float = 0
    so3_quat_eps_float64: float = 0
    so3_hat_eps_float64: float = 0
    se3_hat_eps_float64: float = 0

    def __init__(self):
        self.reset()

    def get_eps(self, ltype: str, attr: str, dtype: torch.dtype) -> float:
        _CHECK_DTYPE_SUPPORTED(dtype)
        attr_name = f"{ltype}_{attr}_eps_{str(dtype)[6:]}"
        return getattr(self, attr_name)

    def reset(self) -> None:
        self.so3_near_pi_eps_float32 = 1e-2
        self.so3_near_zero_eps_float32 = 1e-2
        self.so3_d_near_zero_eps_float32 = 2e-1
        self.so3_matrix_eps_float32 = 4e-4
        self.so3_quat_eps_float32 = 2e-4
        self.so3_hat_eps_float32 = 5e-6
        self.se3_hat_eps_float32 = 5e-6
        self.so3_near_pi_eps_float64 = 1e-7
        self.so3_near_zero_eps_float64 = 5e-3
        self.so3_d_near_zero_eps_float64 = 1e-2
        self.so3_matrix_eps_float64 = 1e-6
        self.so3_quat_eps_float64 = 5e-7
        self.so3_hat_eps_float64 = 5e-7
        self.se3_hat_eps_float64 = 5e-7


_TORCHLIE_GLOBAL_OPTIONS = _TorchLieOptions()


def set_global_options(options: Dict[str, Any]) -> None:
    for k, v in options.items():
        if not hasattr(_TORCHLIE_GLOBAL_OPTIONS, k):
            raise ValueError(f"{k} is not a valid global option for torchlie.")
        setattr(_TORCHLIE_GLOBAL_OPTIONS, k, v)


def reset_global_options() -> None:
    _TORCHLIE_GLOBAL_OPTIONS.reset()
